# ----------------------------------------------------------------------------
# Copyright 2015-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""
Main neon run script.

This script will parse a YAML formatted model configuration file, instantiate the model, and run
a fit.

To run this script, first activate the virtualenv and then run
"python <path to script>/neon <yaml config file> [options]"

Some of the commonly used arguments to this script are:

    -b | --backend [ gpu | cpu | mkl ] : backend to use (gpu backend works on Pascal,
                                         Maxwell and Kepler)

    -e | --epochs [ epochs ] : number of epochs to run fit

    -d | --datatype [ f16 | f32 | f64 ] : floating point data precision to use, note 64-bit
                                            float only works with cpu backend

    -l | --log [ LOGFILE ] : log file for messages, use the -v option to control logging level, if
                             this option is not set then log messages will only go to the screen

    -v[vv] | --verbose : verbosity level

Args that start with '--' (eg. --data_dir) can also be set in a config file (~/nervana/neon.cfg or
specified via -c) by using .ini or .yaml-style syntax (eg. data_dir=value). If an arg is specified
in more than one place, command-line values override config file values which override defaults.

For details of all the command line arguments run this script with the --help option.
"""
import sys
import yaml

from neon import logger
from neon.backends import gen_backend
from neon.callbacks.callbacks import Callbacks
import neon.data
from neon.models import Model
from neon.util.argparser import NeonArgparser
from neon.util.yaml_parse import create_objects


def parse_args():
    """
    Parse command line arguments

    Returns:
        tuple: Contains YAML elements, command line arguments, backend name,
               number of epochs and batch size.
    """
    # setup the arg parser
    parser = NeonArgparser(__doc__)
    parser.add_yaml_arg()

    # parse the cmd line args
    args = parser.parse_args(gen_be=False)

    # load yaml
    yaml_str = args.yaml_file.read()
    root_yaml = yaml.safe_load(yaml_str)

    batch_size = root_yaml['batchsize'] if 'batchsize' in root_yaml else args.batch_size

    if any("--backend" in ag or "-b" in ag for ag in sys.argv):
        # command line overrides yaml setting
        be_name = args.backend
    else:
        be_name = root_yaml['backend'] if 'backend' in root_yaml else 'cpu'

    # command line will override epochs in yaml
    # epochs has default in parser so check argv
    if any("--epochs" in ag or "-e" in ag for ag in sys.argv):
        num_epochs = args.epochs
    else:
        num_epochs = root_yaml['epochs'] if 'epochs' in root_yaml else 1

    gen_backend(backend=be_name,
                rng_seed=args.rng_seed,
                device_id=args.device_id,
                batch_size=batch_size,
                datatype=args.datatype,
                stochastic_round=args.rounding)

    return root_yaml, args, be_name, num_epochs, batch_size


def load_data(data_dir=".", backend_obj=None):
    """
    Load the specified dataset.

    Arguments:
        data_dir (str, optional): Local directory in which to check for and save newly downloaded
                                  datasets.  Defaults to current directory
    Returns:
        tuple: Contains data iterator objects for training and test datasets.
    """

    if root_yaml['dataset']['name'].upper() == 'I1K':
        try:
            import os
            import numpy as np
            from neon.data.dataloader_transformers import OneHot, TypeCast, BGRMeanSubtract
            from neon.data.aeon_shim import AeonDataLoader

            train_config = root_yaml['dataset']['train_config']
            test_config = root_yaml['dataset']['test_config']
            train_config['manifest_filename'] = os.path.join(data_dir,
                                                             'i1k-extracted',
                                                             'train-index.csv')

            test_config['manifest_filename'] = os.path.join(data_dir,
                                                            'i1k-extracted',
                                                            'val-index.csv')

            for config in (train_config, test_config):
                config['type'] = 'image,label'
                config['minibatch_size'] = backend_obj.bsz
                config['macrobatch_size'] = backend_obj.bsz * 12
                config['cache_directory'] = os.path.join(data_dir, 'i1k-cache')

            def wrap_dataloader(dl):
                dl = OneHot(dl, index=1, nclasses=1000)
                dl = TypeCast(dl, index=0, dtype=np.float32)
                dl = BGRMeanSubtract(dl, index=0)
                return dl

            train = wrap_dataloader(AeonDataLoader(train_config, backend_obj))
            test = wrap_dataloader(AeonDataLoader(test_config, backend_obj))

        except (OSError, IOError, ValueError) as err:
            logger.error(err)
            sys.exit(0)
    else:
        if 'dataset' not in root_yaml:
            raise ValueError('dataset not specified in configuration file')
        dataset = getattr(neon.data, root_yaml['dataset']['name'].upper())()
        dataiters = dataset.gen_iterators()
        train = dataiters['train']
        test = dataiters['valid']
    return train, test


if __name__ == "__main__":
    """
    Train and test the specified model.
    """
    root_yaml, args, be_name, num_epochs, batch_size = parse_args()
    model, cost, optim = create_objects(root_yaml,
                                        be_type=be_name,
                                        batch_size=batch_size,
                                        rng_seed=args.rng_seed,
                                        device_id=args.device_id,
                                        default_dtype=args.datatype,
                                        stochastic_rounding=args.rounding)

    if args.model_file:
        model.load_params(args.model_file)
    train, test = load_data(data_dir=args.data_dir, backend_obj=model.be)
    # configure callbacks
    callbacks = Callbacks(model, eval_set=test, **args.callback_args)

    if args.verbose > 2:
        from neon.util.display_information import display_platform_information
        from neon.util.display_information import display_cpu_information
        from neon.util.display_information import display_model_params
        display_platform_information()
        display_cpu_information()
        display_model_params(args, root_yaml)

    if (args.profile is True):
        from neon.benchmark import Benchmark

        inference = args.profile_inference
        model = Model(layers=model.layers, optimizer=optim)
        model.initialize(train, cost=cost)
        b = Benchmark(model)
        if (args.profiling_method == 'time'):
            res = b.time(train, inference=inference, niterations=args.profile_iterations)
            b.print_stats(res, nskip=args.profile_iter_skip)
    else:
        model.fit(train, optimizer=optim, num_epochs=num_epochs, cost=cost, callbacks=callbacks)



from pycuda.tools import context_dependent_memoize
from neon.backends.cuda_templates import _ew_types
from neon.backends.layer_gpu import _magic64
from neon.backends.util.source_module import SourceModule

"""
CUDA kernels for convolutional layers. This code will return one of 3 kernels:
fprop, bprop, or update. Data format is expected to be (from least to most
contiguous dimension):

I (input activations): (C, H, W, N)
F (filters): (C, R, S, K)
O (output activations): (K, P, Q, N)

Where:
    N = batch size
    C = number of input feature maps/channels
    K = number of output feature maps
    H, W = height/width of input feature map
    R, S = height/width of filter
    P, Q = height/width of output feature map

Currently 3D convolution is not supported so it is expected that (D == T == M == 1)
"""


@context_dependent_memoize
def _get_conv_kernel(dtype, filter_size, bsum, operation, filter_bounds_check=False, debug=False):
    """
    Builds the convolution kernel for a specified filter size.

    Arguments:
        dtype (np.dtype): The data type which the kernel will operate on.
        filter_size (int): Total number of elements per filter (R * S)
        bsum (boolean): If set to true, kernel will include code to compute
            batch sum during fprop
        operation (string): Determines which kernel to build. options follow:
            'fprop': Forward propagation of activations.
            'bprop': Backward propagation of error.
            'update': Computes gradients for filter weights based on error and inputs.
        filter_bounds_check (boolean): Checks if filter weight is in bounds when K is
            not a multiple of 32.
        debug (boolean): When set to true, kernels will be compiled with debug symbols.
    """
    assert operation in ["fprop", "bprop", "update"]
    if operation == "fprop" or operation == "update":
        lut_code = r"""
    if(tid < 32)
    {
        int rs = tid;
        int base_x, base_y;

        base_x = output_pixel_x * stride_w - padding_w;
        base_y = output_pixel_y * stride_h - padding_h;

        unsigned int mask = (1 << tid) - 1;

        while(rs < FILTER_SIZE)
        {
            int filter_x, filter_y;
            _idiv_magic32(rs, magic_s, shift_s, S, filter_y, filter_x);

            int index_x = base_x + filter_x * dilation_w;
            int index_y = base_y + filter_y * dilation_h;

            //Check if the index is valid
            int in_bounds = (index_x >= 0 && index_x < W && index_y >= 0 && index_y < H);
            unsigned int threads_in_bounds = __ballot(in_bounds);

            //Store lookup table entry
            if(in_bounds)
            {
                int2 lut_entry;
                lut_entry.x = ((index_y * W + index_x) * N) >> 2;
                lut_entry.y = (rs * K) >> 2;

                int index = lut_size_local + __popc(threads_in_bounds & mask);
                lookup_table[index] = lut_entry;
            }

            lut_size_local += __popc(threads_in_bounds);

            rs += 32;
        }
    }
"""
    elif operation == "bprop":
        lut_code = r"""
    if(tid < 32)
    {
        int rs = tid;
        int base_q, base_p;

        base_q = output_pixel_x - ((S - 1) * dilation_w - padding_w);
        base_p = output_pixel_y - ((R - 1) * dilation_h - padding_h);

        unsigned int mask = (1 << tid) - 1;

        while(rs < FILTER_SIZE)
        {
            int filter_x, filter_y;
            _idiv_magic32(rs, magic_s, shift_s, S, filter_y, filter_x);

            int index_q = base_q + filter_x * dilation_w;
            int index_p = base_p + filter_y * dilation_h;

            //Check if the index is valid
            int in_bounds = (((index_q % stride_w) | (index_p % stride_h)) == 0);
            index_q /= stride_w;
            index_p /= stride_h;
            in_bounds = in_bounds && (index_q >= 0 && index_q < W
                                      && index_p >= 0 && index_p < H);
            unsigned int threads_in_bounds = __ballot(in_bounds);

            //Store lookup table entry
            if(in_bounds)
            {
                int2 lut_entry;
                lut_entry.x = (((index_p * W) + index_q) * N) >> 2;
                lut_entry.y = (rs * K) >> 2;

                int index = lut_size_local + __popc(threads_in_bounds & mask);
                lookup_table[index] = lut_entry;
            }

            lut_size_local += __popc(threads_in_bounds);

            rs += 32;
        }
    }
"""
    if bsum:
        bsum_code = r"""
            float local_bsum = result[q_offset].f[0] + result[q_offset].f[1] +
                               result[q_offset].f[2] + result[q_offset].f[3];
            atomicAdd(&bsum[filter_id], local_bsum);
"""
    else:
        bsum_code = ""

    if operation == "update":
        a_name = "image"
        b_name = "error"
    else:
        if operation == "fprop":
            a_name = "image"
            b_name = "filter"
        elif operation == "bprop":
            a_name = "error"
            b_name = "filter"

    if filter_bounds_check:
        filter_load_cond = "int filter_load_in_bounds = (((filter_id + threadIdx.x) << 2) < K);"
        check_filter_cond = "(!filter_load_in_bounds) ? make_float4(0, 0, 0, 0) :"
    else:
        filter_load_cond = ""
        check_filter_cond = ""

    header_code = r"""
#define TILE_DIM            32
#define ITEMS_PER_THREAD    4
#define THREADS_DIM         8

#define REG_TILE_X          4
#define REG_TILE_Y          4
#define THREADS_DIM_X       8
#define THREADS_DIM_Y       8
#define SM_TILE_X           (REG_TILE_X * THREADS_DIM_X)
#define SM_TILE_Y           (REG_TILE_Y * THREADS_DIM_Y)

#define NUM_ROWS            8
#define FILTER_SIZE         %(filter_size)s
#define MAGIC_FILTER_SIZE   %(magic_filter_size)s
#define SHIFT_FILTER_SIZE   %(shift_filter_size)s

typedef union Matrix {
    %(type)s4 f4;
    %(type)s f[4];
} Matrix;

__device__ inline void _idiv_fast(int numerator, int denominator, float rcp,
                                 int& result, int& remainder)
{
    result = (int)((float)numerator * rcp);
    remainder = numerator - (result * denominator);
    result = (remainder >= denominator) ? (result + 1) : result;
    remainder = (remainder >= denominator) ? (remainder - denominator) : remainder;
}

__device__ inline void _idiv_magic(int numerator, unsigned int magic, unsigned int shift,
                                   int denominator, int& result, int& remainder)
{
    if(magic == 1)
    {
        result = numerator >> shift;
    }
    else
    {
        unsigned long long res64 = numerator * (unsigned long long)magic;
        result = ((int)(res64 >> 32) >> shift);
    }
    remainder = numerator - (result * denominator);
}

__device__ inline void _idiv_magic32(int numerator, unsigned int magic, unsigned int shift,
                                     int denominator, int& result, int& remainder)
{
    if(magic == 1)
    {
        result = numerator >> shift;
    }
    else
    {
        result = ((numerator * magic) >> shift);
    }
    remainder = numerator - (result * denominator);
}

//Note: N and K must be multiples of 4
//blockIdx.x is gemm tile id (K dimension) and output pixel id
//blockIdx.y is gemm tile id (N dimension)
//threadIdx.x is gemm tile offset (K dimension)
//threadIdx.y is gemm tile offset (N dimension)
__global__ void conv_%(operation)s(
                           %(type)s alpha, %(type)s beta,
                           Matrix *I, Matrix *F, Matrix *O, float* bsum,
                           int C, int D, int H, int W, int N,
                           int T, int R, int S, int K,
                           int M, int P, int Q,
                           int stride_w, int stride_h, int padding_w, int padding_h,
                           int dilation_w, int dilation_h,
                           int input_channel_size, int filter_channel_size,
                           int output_filter_size,
                           int output_pixels, int grid_p, int grid_q,
                           unsigned int magic_pq, unsigned int shift_pq,
                           unsigned int magic_q, unsigned int shift_q,
                           unsigned int magic_s, unsigned int shift_s)

"""
    code = r"""
{
    __shared__ int2 lookup_table[FILTER_SIZE];
    __shared__ int lut_size;
    __shared__ Matrix %(a_name)s_data[NUM_ROWS][THREADS_DIM_X];
    __shared__ Matrix %(b_name)s_data[NUM_ROWS][THREADS_DIM_Y];

    int lut_size_local = 0;

    //TODO: Use square access pattern to image data to increase cache hits
    int output_pixel, image_id;
    _idiv_magic(blockIdx.x, magic_pq, shift_pq, output_pixels, image_id, output_pixel);
    image_id = (image_id * blockDim.x);

    //Zig zag along x axis to increase cache hits
    int temp_x, temp_y;
    _idiv_magic(output_pixel, magic_q, shift_q, Q, temp_y, temp_x);
    int output_pixel_x = (temp_y & 1) ? (Q - temp_x - 1) : temp_x;
    int output_pixel_y = temp_y;
    output_pixel = output_pixel_x + (output_pixel_y * Q);

    int filter_id = blockIdx.y * blockDim.y;
    int tid = threadIdx.x + threadIdx.y * blockDim.x;

    //Offset buffers based on thread id
    I = &(I[image_id  + threadIdx.x]);
    F = &(F[filter_id + threadIdx.x]);

    %(filter_load_cond)s

    //Compute lookup table for filter/image data
%(lut_code)s

    if(tid == 0)
    {
        lut_size = lut_size_local;
    }

    __syncthreads();

    lut_size_local = lut_size;
    Matrix result[REG_TILE_Y] = {0};
    output_pixel = (output_pixel * N) >> 2;
    if(lut_size_local > 0)
    {
        //Evaluate gemm with outer product dimensions N, K and inner product CRS
        int CRS = lut_size_local * C;

        //Compute magic numbers for division by lut_size
        float reciprocal = 1.0f / (float)lut_size_local;

        //Initialize shared mem for first block
        int crs = CRS %% NUM_ROWS;
        crs = (crs == 0) ? 8 : crs;

        int c, rs;
        _idiv_fast(CRS - threadIdx.y - 1, lut_size_local, reciprocal, c, rs);

        int2 lut_entry = ((threadIdx.y & 7) >= crs) ? make_int2(0, 0) : lookup_table[rs];
        %(a_name)s_data[threadIdx.y][threadIdx.x].f4 =
            ((threadIdx.y & 7) >= crs) ? make_float4(0, 0, 0, 0) :
            I[(c * input_channel_size)  + lut_entry.x].f4;
        %(b_name)s_data[threadIdx.y][threadIdx.x].f4 = %(check_filter_cond)s
            ((threadIdx.y & 7) >= crs) ? make_float4(0, 0, 0, 0) :
            F[(c * filter_channel_size) + lut_entry.y].f4;

        //Iterate over entire filter
        for(crs = CRS - crs - 1; crs > 0; crs -= NUM_ROWS)
        {
            __syncthreads();

            #pragma unroll
            for(int i = 0; i < NUM_ROWS; i++)
            {
                Matrix load_row;
                Matrix load_col;

                load_row.f4 = %(a_name)s_data[i][threadIdx.x].f4;
                load_col.f4 = %(b_name)s_data[i][threadIdx.y].f4;

                //Accumulate product
                #pragma unroll
                for(int q_offset = 0; q_offset < REG_TILE_Y; q_offset++)
                {
                    #pragma unroll
                    for(int p_offset = 0; p_offset < REG_TILE_X; p_offset++)
                    {
                        result[q_offset].f[p_offset] += (load_row.f[p_offset] *
                                                         load_col.f[q_offset]);
                    }
                }
            }

            __syncthreads();

            //Load new image data and filter weights
            _idiv_fast(crs - threadIdx.y, lut_size_local, reciprocal, c, rs);

            lut_entry = lookup_table[rs];
            %(a_name)s_data[threadIdx.y][threadIdx.x].f4 =
                I[(c * input_channel_size)  + lut_entry.x].f4;
            %(b_name)s_data[threadIdx.y][threadIdx.x].f4 =
                %(check_filter_cond)s F[(c * filter_channel_size) + lut_entry.y].f4;
        }

        __syncthreads();

        //Accumulate product for last iteration
        #pragma unroll
        for(int i = 0; i < NUM_ROWS; i++)
        {
            Matrix load_row;
            Matrix load_col;

            load_row.f4 = %(a_name)s_data[i][threadIdx.x].f4;
            load_col.f4 = %(b_name)s_data[i][threadIdx.y].f4;

            //Accumulate product
            #pragma unroll
            for(int q_offset = 0; q_offset < REG_TILE_Y; q_offset++)
            {
                #pragma unroll
                for(int p_offset = 0; p_offset < REG_TILE_X; p_offset++)
                {
                    result[q_offset].f[p_offset] += (load_row.f[p_offset] * load_col.f[q_offset]);
                }
            }
        }
    }

    //Store result
    filter_id = (filter_id + threadIdx.y) << 2;
    if(filter_id < K)
    {
        image_id += threadIdx.x;

        #pragma unroll
        for(int q_offset = 0; q_offset < 4; q_offset++)
        {
            if(filter_id < K)
            {
                int out_index = (filter_id * output_filter_size) + output_pixel + image_id;
                %(bsum_code)s

                Matrix cur_value = {0};
                if(beta > 0.0f)
                {
                    cur_value.f4 = O[out_index].f4;
                }

                result[q_offset].f[0] = (result[q_offset].f[0] * alpha) + (cur_value.f[0] * beta);
                result[q_offset].f[1] = (result[q_offset].f[1] * alpha) + (cur_value.f[1] * beta);
                result[q_offset].f[2] = (result[q_offset].f[2] * alpha) + (cur_value.f[2] * beta);
                result[q_offset].f[3] = (result[q_offset].f[3] * alpha) + (cur_value.f[3] * beta);

                O[out_index].f4 = result[q_offset].f4;
            }
            filter_id++;
        }
    }
}
"""

    update_code = r"""
{
    __shared__ Matrix %(a_name)s_data[TILE_DIM / 4][THREADS_DIM * 4 + 4];
    __shared__ Matrix %(b_name)s_data[TILE_DIM / 4][THREADS_DIM * 4 + 4];

    //TODO: Use square access pattern to image data to increase cache hits
    int output_pixel, filter_id;
    _idiv_magic(blockIdx.x, magic_pq, shift_pq, output_pixels, filter_id, output_pixel);
    filter_id = filter_id * TILE_DIM;
    int load_filter_id = filter_id + threadIdx.y;

    int filter_pixel_id = blockIdx.y * TILE_DIM;

    //TODO: Zig zag along x axis to increase cache hits
    int output_pixel_x, output_pixel_y;
    _idiv_magic(output_pixel, magic_q, shift_q, grid_q, output_pixel_y, output_pixel_x);

    //Compute input image and filter offsets for this pixel
    int c, rs;
    int crs = filter_pixel_id + threadIdx.y;
    _idiv_magic(crs, MAGIC_FILTER_SIZE, SHIFT_FILTER_SIZE, FILTER_SIZE, c, rs);

    int filter_x, filter_y;
    _idiv_magic32(rs, magic_s, shift_s, S, filter_y, filter_x);

    int output_pixel_x_save = output_pixel_x;
    for(; output_pixel_y < P; output_pixel_y += grid_p)
    {
        for(output_pixel_x = output_pixel_x_save; output_pixel_x < Q; output_pixel_x += grid_q)
        {
            int base_x = output_pixel_x * stride_w - padding_w + filter_x * dilation_w;
            int base_y = output_pixel_y * stride_h - padding_h + filter_y * dilation_h;
            int crs_in_bounds = (c < C) && (base_x >= 0) && (base_x < W) &&
                                (base_y >= 0) && (base_y < H);
            int input_pixel = W * base_y + base_x;
            output_pixel = output_pixel_x + (Q * output_pixel_y);

            //Pre-multiply offset to simplify indexing
            input_pixel = (input_pixel * N) >> 2;
            output_pixel = (output_pixel * N) >> 2;

            //Evaluate gemm with outer product dimensions N, K and inner product CRS
            Matrix result[ITEMS_PER_THREAD] = {0};

            //Load image data and transpose into shared mem
            //TODO: pad shared memory to avoid bank conflicts
            Matrix buffer;
            buffer.f4 = crs_in_bounds ?
                        I[(c * input_channel_size) + input_pixel + threadIdx.x].f4 :
                        make_float4(0, 0, 0, 0);
            %(a_name)s_data[threadIdx.x][ 0 | threadIdx.y >> 2].f[threadIdx.y & 3] = buffer.f[0];
            %(a_name)s_data[threadIdx.x][ 8 | threadIdx.y >> 2].f[threadIdx.y & 3] = buffer.f[1];
            %(a_name)s_data[threadIdx.x][16 | threadIdx.y >> 2].f[threadIdx.y & 3] = buffer.f[2];
            %(a_name)s_data[threadIdx.x][24 | threadIdx.y >> 2].f[threadIdx.y & 3] = buffer.f[3];

            //Load error data and transpose into shared mem
            buffer.f4 = (load_filter_id < K) ?
                        F[(load_filter_id * output_filter_size) + output_pixel + threadIdx.x].f4 :
                        make_float4(0, 0, 0, 0);
            %(b_name)s_data[threadIdx.x][ 0 | threadIdx.y >> 2].f[threadIdx.y & 3] = buffer.f[0];
            %(b_name)s_data[threadIdx.x][ 8 | threadIdx.y >> 2].f[threadIdx.y & 3] = buffer.f[1];
            %(b_name)s_data[threadIdx.x][16 | threadIdx.y >> 2].f[threadIdx.y & 3] = buffer.f[2];
            %(b_name)s_data[threadIdx.x][24 | threadIdx.y >> 2].f[threadIdx.y & 3] = buffer.f[3];

            //Iterate over entire minibatch
            for(int n = threadIdx.x + (TILE_DIM >> 2); n < (N >> 2); n += (TILE_DIM >> 2))
            {
                __syncthreads();

                #pragma unroll
                for(int i = 0; i < (TILE_DIM >> 2); i++)
                {
                    Matrix row_image;
                    Matrix row_error;

                    row_image.f4 =
                        %(a_name)s_data[i][((threadIdx.y & 3) << 3) | threadIdx.y >> 2].f4;
                    row_error.f4 =
                        %(b_name)s_data[i][((threadIdx.y & 3) << 3) | threadIdx.x].f4;

                    //Accumulate product
                    #pragma unroll
                    for(int q_offset = 0; q_offset < ITEMS_PER_THREAD; q_offset++)
                    {
                        #pragma unroll
                        for(int p_offset = 0; p_offset < ITEMS_PER_THREAD; p_offset++)
                        {
                            result[p_offset].f[q_offset] +=
                                (row_image.f[p_offset] * row_error.f[q_offset]);
                        }
                    }
                }

                __syncthreads();

                //Load image data and transpose into shared mem
                buffer.f4 = crs_in_bounds ?
                    I[(c * input_channel_size) + input_pixel + n].f4 :
                    make_float4(0, 0, 0, 0);
                %(a_name)s_data[threadIdx.x][ 0 | threadIdx.y >> 2].f[threadIdx.y & 3] =
                    buffer.f[0];
                %(a_name)s_data[threadIdx.x][ 8 | threadIdx.y >> 2].f[threadIdx.y & 3] =
                    buffer.f[1];
                %(a_name)s_data[threadIdx.x][16 | threadIdx.y >> 2].f[threadIdx.y & 3] =
                    buffer.f[2];
                %(a_name)s_data[threadIdx.x][24 | threadIdx.y >> 2].f[threadIdx.y & 3] =
                    buffer.f[3];

                //Load error data and transpose into shared mem
                buffer.f4 = (load_filter_id < K) ?
                    F[(load_filter_id * output_filter_size) + output_pixel + n].f4 :
                    make_float4(0, 0, 0, 0);
                %(b_name)s_data[threadIdx.x][ 0 | threadIdx.y >> 2].f[threadIdx.y & 3] =
                    buffer.f[0];
                %(b_name)s_data[threadIdx.x][ 8 | threadIdx.y >> 2].f[threadIdx.y & 3] =
                    buffer.f[1];
                %(b_name)s_data[threadIdx.x][16 | threadIdx.y >> 2].f[threadIdx.y & 3] =
                    buffer.f[2];
                %(b_name)s_data[threadIdx.x][24 | threadIdx.y >> 2].f[threadIdx.y & 3] =
                    buffer.f[3];
            }

            __syncthreads();

            //Accumulate product for last iteration
            #pragma unroll
            for(int i = 0; i < (TILE_DIM >> 2); i++)
            {
                Matrix row_image;
                Matrix row_error;

                row_image.f4 = %(a_name)s_data[i][((threadIdx.y & 3) << 3) | threadIdx.y >> 2].f4;
                row_error.f4 = %(b_name)s_data[i][((threadIdx.y & 3) << 3) | threadIdx.x].f4;

                //Accumulate product
                #pragma unroll
                for(int q_offset = 0; q_offset < ITEMS_PER_THREAD; q_offset++)
                {
                    #pragma unroll
                    for(int p_offset = 0; p_offset < ITEMS_PER_THREAD; p_offset++)
                    {
                        result[p_offset].f[q_offset] +=
                            (row_image.f[p_offset] * row_error.f[q_offset]);
                    }
                }
            }

            //Reduce result between threads in warp
            Matrix outbound;
            int warp_y = threadIdx.y & 3;
            int warp_id = threadIdx.x + (threadIdx.y << 3);
            buffer.f4 = (warp_y == 0) ? result[0].f4 :
                        (warp_y == 1) ? result[1].f4 :
                        (warp_y == 2) ? result[2].f4 :
                        result[3].f4;

            outbound.f4 = (warp_y == 0) ? result[3].f4 :
                          (warp_y == 1) ? result[0].f4 :
                          (warp_y == 2) ? result[1].f4 :
                          result[2].f4;
            buffer.f[0] += __shfl(outbound.f[0], warp_id + 8);
            buffer.f[1] += __shfl(outbound.f[1], warp_id + 8);
            buffer.f[2] += __shfl(outbound.f[2], warp_id + 8);
            buffer.f[3] += __shfl(outbound.f[3], warp_id + 8);

            outbound.f4 = (warp_y == 0) ? result[2].f4 :
                          (warp_y == 1) ? result[3].f4 :
                          (warp_y == 2) ? result[0].f4 :
                          result[1].f4;
            buffer.f[0] += __shfl(outbound.f[0], warp_id + 16);
            buffer.f[1] += __shfl(outbound.f[1], warp_id + 16);
            buffer.f[2] += __shfl(outbound.f[2], warp_id + 16);
            buffer.f[3] += __shfl(outbound.f[3], warp_id + 16);

            outbound.f4 = (warp_y == 0) ? result[1].f4 :
                          (warp_y == 1) ? result[2].f4 :
                          (warp_y == 2) ? result[3].f4 :
                          result[0].f4;
            buffer.f[0] += __shfl(outbound.f[0], warp_id + 24);
            buffer.f[1] += __shfl(outbound.f[1], warp_id + 24);
            buffer.f[2] += __shfl(outbound.f[2], warp_id + 24);
            buffer.f[3] += __shfl(outbound.f[3], warp_id + 24);

            //Store result
            int idx_filter_id = filter_id + (threadIdx.x << 2);
            if(idx_filter_id < K && crs_in_bounds)
            {
                int out_index = (c * filter_channel_size) + (((rs * K) + (idx_filter_id)) >> 2);

                atomicAdd(&O[out_index].f[0], buffer.f[0]);
                atomicAdd(&O[out_index].f[1], buffer.f[1]);
                atomicAdd(&O[out_index].f[2], buffer.f[2]);
                atomicAdd(&O[out_index].f[3], buffer.f[3]);
            }
        }
    }
}
"""
    if operation == "update":
        code = header_code + update_code
    else:
        code = header_code + code

    magic = _magic64(filter_size)

    code = code % {
        "filter_size": filter_size,
        "magic_filter_size": magic[0],
        "shift_filter_size": magic[1],
        "type": _ew_types[dtype]["type"],
        "lut_code": lut_code,
        "bsum_code": bsum_code,
        "operation": operation,
        "a_name": a_name,
        "b_name": b_name,
        "filter_load_cond": filter_load_cond,
        "check_filter_cond": check_filter_cond
    }

    options = ["--use_fast_math"]
    if debug and operation == "bprop":
        options = options + ["-g", "-G"]
    module = SourceModule(code, options=options)

    kernel = module.get_function("conv_" + operation)
    kernel.prepare("ffPPPPIIIIIIIIIIIIIIIIIIIIIIIIIIIIII")
    kernel.name = "conv_" + operation
    return kernel



# ----------------------------------------------------------------------------
# Copyright 2017 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""
MKL backend layers
"""
from __future__ import division
import numpy as np
from ctypes import c_longlong, c_float
from neon.backends.layer_cpu import ConvLayer, DeconvLayer, PoolLayer


class ConvLayerMKL(ConvLayer):

    """
    MKL based ConvLayer
    """
    def __init__(self, lib, dtype,
                 N, C, K,
                 D=1, H=1, W=1,
                 T=1, R=1, S=1,
                 pad_d=0, pad_h=0, pad_w=0,
                 str_d=1, str_h=1, str_w=1,
                 dil_d=1, dil_h=1, dil_w=1):

        super(ConvLayerMKL, self).__init__(lib, dtype,
                                           N, C, K, D, H, W, T, R, S,
                                           pad_d, pad_h, pad_w, str_d, str_h, str_w,
                                           dil_d, dil_h, dil_w)
        self.dnnPrimitives = np.zeros((1, 52), dtype=np.uint64)
        self.init_f = 0  # forward init flag
        self.init_bd = 0  # backward data
        self.init_bw = 0  # backward weight
        self.dilated = any(d != 1 for d in self.dilation)
        self.is_mklop = True
        if D != 1 or T != 1 or dil_d != 1:
            self.is_mklop = False

    def xprop_conv(self, I, F, O, X=None, bias=None, bsum=None, alpha=1.0, beta=0.0,
                   relu=False, brelu=False, slope=0.0, backward=False, layer_op=None):

        if layer_op is None:
            layer_op = self

        if not self.get_is_mklop():
            I.backend.convert(I)
            F.backend.convert(F)
            I.clean_mkl()
            F.clean_mkl()
            super(ConvLayerMKL, self).xprop_conv(
                I, F, O, X, bias, bsum, alpha, beta, relu, brelu, slope, backward)
            return

        if X is None:
            X = O

        C, D, H, W, N = self.dimI
        C, T, R, S, K = self.dimF
        K, M, P, Q, N = self.dimO
        pad_d, pad_h, pad_w = self.padding
        str_d, str_h, str_w = self.strides
        dil_d, dil_h, dil_w = self.dilation
        primitives = c_longlong(self.dnnPrimitives.ctypes.data)
        bias_prim = c_longlong(0)
        if bias is not None:
            bias_prim = bias.get_prim()

        mkl_res = 0
        if not backward:
            mkl_res = I.backend.mklEngine.Conv_forward(
                I.get_prim(), O.get_prim(), F.get_prim(), bias_prim, primitives, self.init_f,
                N, C, H, W, R, S, str_h, str_w, pad_h, pad_w, dil_h, dil_w, K, P, Q)
            self.init_f = 1
            O.shape5D = self.dimO
        else:
            I.backend.mklEngine.Conv_bwdData(
                I.get_prim(), O.get_prim(), F.get_prim(), primitives,
                N, C, H, W, R, S, str_h, str_w, pad_h, pad_w,
                dil_h, dil_w, K, P, Q, self.init_bd, c_float(beta))
            O.shape5D = self.dimI
            self.init_bd = 1
        if mkl_res != 0:
            super(ConvLayerMKL, self).xprop_conv(
                I, F, O, X, bias, bsum, alpha, beta, relu, brelu, slope, backward)
            I.clean_mkl()
            O.clean_mkl()
            layer_op.set_not_mklop()
            return

    # grad_bias is added for convolution layer with bias
    def update_conv(self, I, E, U, alpha=1.0, beta=0.0,  grad_bias=None, layer_op=None):

        if not self.get_is_mklop():
            I.backend.convert(I)
            I.clean_mkl()
            E.backend.convert(E)
            E.clean_mkl()
            super(ConvLayerMKL, self).update_conv(I, E, U, alpha, beta, grad_bias, layer_op)
            return

        # not deal with alpha, beta yet
        K, M, P, Q, N = self.dimO
        C, D, H, W, N = self.dimI
        C, T, R, S, K = self.dimF
        pad_d, pad_h, pad_w = self.padding
        str_d, str_h, str_w = self.strides
        dil_d, dil_h, dil_w = self.dilation
        primitives = c_longlong(self.dnnPrimitives.ctypes.data)
        bias_prim = c_longlong(0)
        if grad_bias is not None:
            bias_prim = grad_bias.get_prim()
        I.backend.mklEngine.Conv_bwdFilter(
            I.get_prim(), E.get_prim(), U.get_prim(), bias_prim, primitives,
            N, C, H, W, R, S, str_h, str_w, pad_h, pad_w,
            dil_h, dil_w, K, P, Q, self.init_bw, self.init_bd)
        self.init_bw = 1


class DeconvLayerMKL(DeconvLayer):

    """
    DeconvLayer parameter object.
    This then is passed as an argument to all the convolution operations.

    N: Number of images in mini-batch
    C: Number of output feature maps
    K: Number of input feature maps

    P: Height of input
    Q: Width of input

    D: Depth  of output image
    H: Height of output image
    W: Width  of output image

    T: Depth  of filter kernel
    R: Height of filter kernel
    S: Width  of filter kernel

    padding: amount of zero-padding around the given edge
    strides: factor to step the filters by in a given direction
    dilation: dilation factor for each dimension
    """

    def xprop_conv(self, I, F, O, X=None, bias=None, bsum=None, alpha=1.0, beta=0.0,
                   relu=False, brelu=False, slope=0.0, backward=False, layer_op=None):

        I.backend.convert(I)
        I.clean_mkl()
        O.backend.convert(O)
        O.clean_mkl()
        super(DeconvLayerMKL, self).xprop_conv(I, F, O, X, bias, bsum, alpha, beta,
                                               relu, brelu, slope, backward, layer_op)

    def update_conv(self, I, E, U, alpha=1.0, beta=0.0, grad_bias=None, layer_op=None):

        I.backend.convert(I)
        I.clean_mkl()

        E.backend.convert(E)
        E.clean_mkl()

        super(DeconvLayerMKL, self).update_conv(I, E, U, alpha, beta, grad_bias, layer_op)


class PoolLayerMKL(PoolLayer):

    """
    PoolLayer parameter object.
    This then is passed as an argument to all pooling kernels.

    op: max, avg, l2 pooling
    N: Number of images in mini-batch

    C: Number of input feature maps
    D: Depth  of input image
    H: Height of input image
    W: Width  of input image

    J: Size of feature map pooling window (maxout n_pieces)
    T: Depth  of pooling window
    R: Height of pooling window
    S: Width  of pooling window

    padding: amount of zero-padding around the given image or feature map edge
    strides: factor to step the window by in a given direction (overlap allowed)

    Leave spatial dimensions at 1 to allow feature map pooling in the fc layers.

    """

    def __init__(self, lib, dtype,
                 op, N, C,
                 D=1, H=1, W=1,
                 J=1, T=1, R=1, S=1,
                 pad_c=0, pad_d=0, pad_h=0, pad_w=0,
                 str_c=None, str_d=None, str_h=None, str_w=None):

        super(PoolLayerMKL, self).__init__(lib, dtype,
                                           op, N, C, D, H, W, J, T, R, S,
                                           pad_c, pad_d, pad_h, pad_w,
                                           str_c, str_d, str_h, str_w)
        self.dnnPrimitives = np.zeros((1, 20), dtype=np.uint64)
        self.initOk_f = 0
        self.initOk_b = 0
        self.is_mklop = True


class ReluLayerMKL(object):
    '''
    Relu layer for MKL backend
    Initialize primitive
    '''
    def __init__(self):
        self.dnnPrimitives = np.zeros((1, 12), dtype=np.uint64)
        self.initOk_b = 0
        self.initOk_f = 0
        self.inputMKL = True
        self.shape5D = None


class BatchNormLayerMKL(object):
    '''
    Relu layer for MKL backend
    Initialize primitive
    '''
    def __init__(self, in_shape):
        self.dnnPrimitives = np.zeros((1, 20), dtype=np.uint64)
        self.init_f = 0
        self.init_b = 0
        self.in_shape = in_shape
        self.shape5D = None


class MergeSumLayerMKL(object):
    '''
    MergeSum for MKL backend
    Initialize primitive
    '''
    def __init__(self, layer_num):

        # declare forward sum, backward sum and temperary tensors used in sum
        self.sum_prim_f = np.zeros((1), dtype=np.uint64)
        self.sum_prim_b = np.zeros((1), dtype=np.uint64)
        self.tensors = np.zeros((4 * layer_num), dtype=np.uint64)
        # self.in_shape = in_shape
        self.shape5D = None
        self.layer_num = layer_num


class MergeBroadcastLayerMKL(object):
    '''
    MergeBroadcast for MKL backend
    '''
    def __init__(self, layer_num):
        # mkl internal memory pointer
        self.primitive = np.zeros((15), dtype=np.uint64)
        self.sum_prim = np.zeros((1),  dtype=np.uint64)

        # temp tensors pointers
        self.tensors_temp = np.zeros((4 * layer_num), dtype=np.uint64)

        self.initOK_f = 0
        self.initOK_b = 0
        self.in_shape5D = None
        self.out_shape5D = None
        self.layer_num = layer_num

        # to record input channels
        self.channels = np.zeros(layer_num, dtype=np.uint64)


class BiBNRNNLayerMKL(object):
    '''
    BiBNRNN for MKL backend
    Initialize parameters
    '''
    def __init__(self, h_buffer_all, h_ff_buffer, W_recur_f, W_recur_b, nsteps, nout):
        self.h_all_contiguous = h_buffer_all.backend.empty_like(h_buffer_all)
        h_all_tmp = self.h_all_contiguous.reshape(nsteps + 2, 2 * nout, -1)
        h_all_contiguous_f_list = [h_all_tmp[step, 0: nout, :] for step in range(nsteps + 2)]
        self.h_prev_contiguous = h_all_contiguous_f_list[0:nsteps]
        self.h_f_contiguous = h_all_contiguous_f_list[1:nsteps + 1]

        h_all_contiguous_b_list = [h_all_tmp[step, nout:, :] for step in range(nsteps + 2)]
        self.h_next_contiguous = h_all_contiguous_b_list[2:nsteps + 2]
        self.h_b_contiguous = h_all_contiguous_b_list[1:nsteps + 1]

        self.h_ff_buffer_contiguous = h_buffer_all.backend.empty_like(h_ff_buffer)
        h_ff_tmp = self.h_ff_buffer_contiguous.reshape(nsteps, 2 * nout, -1)
        self.h_ff_f_contiguous = [h_ff_tmp[step, 0:nout, :] for step in range(nsteps)]
        self.h_ff_b_contiguous = [h_ff_tmp[step, nout:, :] for step in range(nsteps)]

        self.error_contiguous = h_buffer_all.backend.empty((2 * nout, nsteps * h_ff_tmp.shape[2]))
        error_tmp = self.error_contiguous.reshape(nsteps, 2 * nout, -1)
        self.in_deltas_f = [error_tmp[step, 0:nout, :] for step in range(nsteps)]
        self.prev_in_deltas = self.in_deltas_f[-1:] + self.in_deltas_f[:-1]
        self.in_deltas_b = [error_tmp[step, nout:, :] for step in range(nsteps)]
        self.next_in_deltas = self.in_deltas_b[1:] + self.in_deltas_b[:1]

        self.W_recur_f_T_contiguous = h_buffer_all.backend.empty_like(W_recur_f)
        self.W_recur_b_T_contiguous = h_buffer_all.backend.empty_like(W_recur_b)



#!/usr/bin/env python
# ----------------------------------------------------------------------------
# Copyright 2015-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
from neon import logger as neon_logger
import os
import sys


def get_mkl_lib(device_id=None, verbose=False):
    if sys.platform == 'win32':
        # find *.dll
        current_path = os.path.dirname(os.path.realpath(__file__))
        mkl_engine_path = os.path.join(current_path, os.pardir, 'mklEngine', 'mklml.dll')
        if not os.path.isfile(mkl_engine_path):
            neon_logger.display("mklml.dll not found")
            return 0

        mkl_engine_path = os.path.join(current_path, os.pardir, 'mklEngine', 'mklEngine.dll')
        if not os.path.isfile(mkl_engine_path):
            neon_logger.display("mklEngine.dll not found")
            return 0

        math_engine_path = os.path.join(current_path, os.pardir, 'mklEngine', 'cmath.dll')
        if not os.path.isfile(math_engine_path):
            neon_logger.display("cmath.dll not found")
            return 0

        header_path = os.path.join(os.path.dirname(__file__), 'mklEngine',
                                   'src', 'math_cpu.header')
        if os.path.isfile(header_path):
            neon_logger.display("math_cpu.header not found")
            return 0
        return 1

    elif sys.platform == 'darwin':
        # find *.dylib
        current_path = os.path.dirname(os.path.realpath(__file__))
        mkl_engine_path = os.path.join(current_path, os.pardir, 'mklEngine', 'mklEngine.dylib')
        if not os.path.isfile(mkl_engine_path):
            neon_logger.display("mklEngine.dylib not found")
            return 0

        math_engine_path = os.path.join(current_path, os.pardir, 'mklEngine', 'cmath.dylib')
        if not os.path.isfile(math_engine_path):
            neon_logger.display("cmath.dylib not found")
            return 0

        header_path = os.path.join(os.path.dirname(__file__), 'mklEngine',
                                   'src', 'math_cpu.header')
        if os.path.isfile(header_path):
            neon_logger.display("math_cpu.header not found")
            return 0
        return 1

    else:
        # find *.so
        current_path = os.path.dirname(os.path.realpath(__file__))
        mkl_engine_path = os.path.join(current_path, os.pardir, 'mklEngine', 'mklEngine.so')
        if not os.path.isfile(mkl_engine_path):
            neon_logger.display("mklEngine.so not found")
            return 0

        math_engine_path = os.path.join(current_path, os.pardir, 'mklEngine', 'cmath.so')
        if not os.path.isfile(math_engine_path):
            neon_logger.display("cmath.so not found")
            return 0

        header_path = os.path.join(os.path.dirname(__file__), 'mklEngine',
                                   'src', 'math_cpu.header')
        if os.path.isfile(header_path):
            neon_logger.display("math_cpu.header not found")
            return 0
        return 1



# Copyright 2014-2016 Nervana Systems Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from builtins import str
import re
import os.path
import subprocess
import pycuda.driver as drv
from pycuda.tools import context_dependent_memoize
from neon import logger as neon_logger
from neon.util.persist import get_cache_dir

# helpful for kernel development
debug = 0

base_dir  = os.path.dirname(__file__)
maxas_dir = os.path.join(base_dir, "kernels", "maxas")
sass_dir  = os.path.join(base_dir, "kernels", "sass")

ptx_dir   = get_cache_dir(['kernels', 'ptx'])
pre_dir   = get_cache_dir(['kernels', 'pre'])
cubin_dir = get_cache_dir(['kernels', 'cubin'])
dump_dir  = get_cache_dir(['kernels', 'dump'])

kernels = {
    "sconv_direct_fprop_128x128": {"threads": 256, "sass": "sconv_xprop_X128_N128", "params": "fprop",  "share": "128*8*2 + 128*8*2 + 10", "args": {"prop": "f"}},
    "sconv_direct_bprop_128x128": {"threads": 256, "sass": "sconv_xprop_X128_N128", "params": "bprop",  "share": "128*8*2 + 128*8*2 + 10", "args": {"prop": "b"}},
    "hconv_direct_fprop_128x128": {"threads": 256, "sass": "hconv_xprop_X128_N128", "params": "fprop",  "share": "128*8*2 + 128*8*2 + 10", "args": {"prop": "f"}},
    "hconv_direct_bprop_128x128": {"threads": 256, "sass": "hconv_xprop_X128_N128", "params": "bprop",  "share": "128*8*2 + 128*8*2 + 10", "args": {"prop": "b"}},

    "sconv_direct_fprop_64x128":  {"threads": 128, "sass": "sconv_xprop_X64_N128",  "params": "fprop",  "share": " 64*8*2 + 128*8*2 + 10", "args": {"prop": "f"}},
    "sconv_direct_bprop_64x128":  {"threads": 128, "sass": "sconv_xprop_X64_N128",  "params": "bprop",  "share": " 64*8*2 + 128*8*2 + 10", "args": {"prop": "b"}},
    "hconv_direct_fprop_64x128":  {"threads": 128, "sass": "hconv_xprop_X64_N128",  "params": "fprop",  "share": " 64*8*2 + 128*8*2 + 10", "args": {"prop": "f"}},
    "hconv_direct_bprop_64x128":  {"threads": 128, "sass": "hconv_xprop_X64_N128",  "params": "bprop",  "share": " 64*8*2 + 128*8*2 + 10", "args": {"prop": "b"}},

    "sconv_direct_fprop_32x128":  {"threads":  64, "sass": "sconv_xprop_X32_N128",  "params": "fprop",  "share": " 32*8*2 + 128*8*2 + 10", "args": {"prop": "f"}},
    "sconv_direct_bprop_32x128":  {"threads":  64, "sass": "sconv_xprop_X32_N128",  "params": "bprop",  "share": " 32*8*2 + 128*8*2 + 10", "args": {"prop": "b"}},
    "hconv_direct_fprop_32x128":  {"threads":  64, "sass": "hconv_xprop_X32_N128",  "params": "fprop",  "share": " 32*8*2 + 128*8*2 + 10", "args": {"prop": "f"}},
    "hconv_direct_bprop_32x128":  {"threads":  64, "sass": "hconv_xprop_X32_N128",  "params": "bprop",  "share": " 32*8*2 + 128*8*2 + 10", "args": {"prop": "b"}},

    "sconv_direct_fprop_128x64":  {"threads": 128, "sass": "sconv_xprop_X128_N64",  "params": "fprop",  "share": "128*8*2 +  64*8*2 + 10", "args": {"prop": "f"}},
    "sconv_direct_bprop_128x64":  {"threads": 128, "sass": "sconv_xprop_X128_N64",  "params": "bprop",  "share": "128*8*2 +  64*8*2 + 10", "args": {"prop": "b"}},
    "hconv_direct_fprop_128x64":  {"threads": 128, "sass": "hconv_xprop_X128_N64",  "params": "fprop",  "share": "128*8*2 +  64*8*2 + 10", "args": {"prop": "f"}},
    "hconv_direct_bprop_128x64":  {"threads": 128, "sass": "hconv_xprop_X128_N64",  "params": "bprop",  "share": "128*8*2 +  64*8*2 + 10", "args": {"prop": "b"}},

    "sconv_direct_fprop_64x64":   {"threads":  64, "sass": "sconv_xprop_X64_N64",   "params": "fprop",  "share": " 64*8*2 +  64*8*2 + 10", "args": {"prop": "f"}},
    "sconv_direct_bprop_64x64":   {"threads":  64, "sass": "sconv_xprop_X64_N64",   "params": "bprop",  "share": " 64*8*2 +  64*8*2 + 10", "args": {"prop": "b"}},
    "hconv_direct_fprop_64x64":   {"threads":  64, "sass": "hconv_xprop_X64_N64",   "params": "fprop",  "share": " 64*8*2 +  64*8*2 + 10", "args": {"prop": "f"}},
    "hconv_direct_bprop_64x64":   {"threads":  64, "sass": "hconv_xprop_X64_N64",   "params": "bprop",  "share": " 64*8*2 +  64*8*2 + 10", "args": {"prop": "b"}},

    "sconv_direct_fprop_64x32":   {"threads": 128, "sass": "xconv_direct_xprop_64x32",  "params": "fprop2", "share": "(32 + 64)*32*2 + 4", "args": {"type":"s","prop":"f"}},
    "sconv_direct_bprop_64x32":   {"threads": 128, "sass": "xconv_direct_xprop_64x32",  "params": "bprop2", "share": "(32 + 64)*32*2 + 4", "args": {"type":"s","prop":"b"}},
    "hconv_direct_fprop_64x32":   {"threads": 128, "sass": "xconv_direct_xprop_64x32",  "params": "fprop2", "share": "(32 + 64)*32*2 + 4", "args": {"type":"h","prop":"f"}},
    "hconv_direct_bprop_64x32":   {"threads": 128, "sass": "xconv_direct_xprop_64x32",  "params": "bprop2", "share": "(32 + 64)*32*2 + 4", "args": {"type":"h","prop":"b"}},
    "sconv_direct_updat_64x32":   {"threads": 128, "sass": "xconv_direct_updat_64x32",  "params": "updat2", "share": "(32 + 64)*33*2 + 8", "args": {"type": "s",}},
    "hconv_direct_updat_64x32":   {"threads": 128, "sass": "xconv_direct_updat_64x32",  "params": "updat2", "share": "(32 + 64)*33*2 + 8", "args": {"type": "h",}},


    "sconv_winograd_2x2_3x3_32x32":   {"threads": 256, "sass": "xconv_winograd_2x2_3x3_32x32",   "params": "fpropw", "share": "512*4*4", "args": {"type": "s"}},
    "hconv_winograd_2x2_3x3_32x32":   {"threads": 256, "sass": "xconv_winograd_2x2_3x3_32x32",   "params": "fpropw", "share": "512*4*4", "args": {"type": "h"}},
    "sconv_winograd_3x3_2x2_32x32":   {"threads": 256, "sass": "xconv_winograd_3x3_2x2_32x32",   "params": "updatw", "share": "(512*4 + 32)*4 + 8", "args": {"type": "s"}},
    "hconv_winograd_3x3_2x2_32x32":   {"threads": 256, "sass": "xconv_winograd_3x3_2x2_32x32",   "params": "updatw", "share": "(512*4 + 32)*4 + 8", "args": {"type": "h"}},

    "sconv_winograd_4x4_3x3_32x32":   {"threads": 640, "sass": "xconv_winograd_4x4_3x3_32x32",   "params": "fpropw4",  "share": "32*36*2*4 + 64 + 8", "args": {"type": "s"}},
    "hconv_winograd_4x4_3x3_32x32":   {"threads": 640, "sass": "xconv_winograd_4x4_3x3_32x32",   "params": "fpropw4",  "share": "32*36*2*4 + 64 + 8", "args": {"type": "h"}},
    "sconv_winograd_4x4_3x3_32x32_X": {"threads": 640, "sass": "xconv_winograd_4x4_3x3_32x32_X", "params": "fpropw4X", "share": "32*36*2*4 + 64 + 8", "args": {"type": "s"}},
    "hconv_winograd_4x4_3x3_32x32_X": {"threads": 640, "sass": "xconv_winograd_4x4_3x3_32x32_X", "params": "fpropw4X", "share": "32*36*2*4 + 64 + 8", "args": {"type": "h"}},
    "sconv_winograd_3x3_4x4_32x32":   {"threads": 640, "sass": "xconv_winograd_3x3_4x4_32x32",   "params": "updatw4",  "share": "32*36*2*4 + 64 + 8", "args": {"type": "s"}},
    "hconv_winograd_3x3_4x4_32x32":   {"threads": 640, "sass": "xconv_winograd_3x3_4x4_32x32",   "params": "updatw4",  "share": "32*36*2*4 + 64 + 8", "args": {"type": "h"}},

    "sconv_winograd_2x2_5x5_32x32":   {"threads": 640, "sass": "xconv_winograd_2x2_5x5_32x32",   "params": "fpropw5",  "share": "32*36*2*4 + 64 + 8", "args": {"type": "s"}},
    "hconv_winograd_2x2_5x5_32x32":   {"threads": 640, "sass": "xconv_winograd_2x2_5x5_32x32",   "params": "fpropw5",  "share": "32*36*2*4 + 64 + 8", "args": {"type": "h"}},


    "sgemm_nn_128x128": {"threads": 256, "sass": "sgemm_nn_128x128", "params": "gemm", "share": "128*8*2 + 128*8*2 + 4"},
    "sgemm_nt_128x128": {"threads": 256, "sass": "sgemm_nt_128x128", "params": "gemm", "share": "128*8*2 + 128*8*2 + 4"},
    "sgemm_tn_128x128": {"threads": 256, "sass": "sgemm_tn_128x128", "params": "gemm", "share": "128*8*2 + 128*8*2 + 4"},
    "hgemm_nn_128x128": {"threads": 256, "sass": "hgemm_nn_128x128", "params": "gemm", "share": "128*8*2 + 128*8*2 + 4"},
    "hgemm_nt_128x128": {"threads": 256, "sass": "hgemm_nt_128x128", "params": "gemm", "share": "128*8*2 + 128*8*2 + 4"},
    "hgemm_tn_128x128": {"threads": 256, "sass": "hgemm_tn_128x128", "params": "gemm", "share": "128*8*2 + 128*8*2 + 4"},

    "sgemm_nn_128x64":  {"threads": 128, "sass": "sgemm_nn_128x64",  "params": "gemm", "share": "128*8*2 +  64*8*2 + 4"},
    "sgemm_tn_128x64":  {"threads": 128, "sass": "sgemm_tn_128x64",  "params": "gemm", "share": "128*8*2 +  64*8*2 + 4"},
    "hgemm_nn_128x64":  {"threads": 128, "sass": "hgemm_nn_128x64",  "params": "gemm", "share": "128*8*2 +  64*8*2 + 4"},
    "hgemm_tn_128x64":  {"threads": 128, "sass": "hgemm_tn_128x64",  "params": "gemm", "share": "128*8*2 +  64*8*2 + 4"},

    "sgemm_nn_128x32":  {"threads": 128, "sass": "sgemm_nn_128x32",  "params": "gemm", "share": "(128*16 + 32)*2 + 32*16*2 + 4"},
    "sgemm_tn_128x32":  {"threads": 128, "sass": "sgemm_tn_128x32",  "params": "gemm", "share": "(128*16 +  0)*2 + 32*16*2 + 4"},
    "hgemm_nn_128x32":  {"threads": 128, "sass": "hgemm_nn_128x32",  "params": "gemm", "share": "(128*16 + 32)*2 + 32*16*2 + 4"},
    "hgemm_tn_128x32":  {"threads": 128, "sass": "hgemm_tn_128x32",  "params": "gemm", "share": "(128*16 +  0)*2 + 32*16*2 + 4"},

    "sgemm_nn_32x128":  {"threads": 128, "sass": "sgemm_nn_32x128",  "params": "gemm", "share": "(32*16 + 32)*2 + (128*16 +  0)*2 + 4"},
    "sgemm_nt_32x128":  {"threads": 128, "sass": "sgemm_nt_32x128",  "params": "gemm", "share": "(32*16 + 32)*2 + (128*16 + 32)*2 + 4"},
    "hgemm_nn_32x128":  {"threads": 128, "sass": "hgemm_nn_32x128",  "params": "gemm", "share": "(32*16 + 32)*2 + (128*16 +  0)*2 + 4"},
    "hgemm_nt_32x128":  {"threads": 128, "sass": "hgemm_nt_32x128",  "params": "gemm", "share": "(32*16 + 32)*2 + (128*16 + 32)*2 + 4"},

    "hgemm_nt_32x32": {"threads": 128, "sass": "hgemm_nt_32x32", "params": "gemm", "share": "32*65*4 + 4" },
    "hgemm_nt_16x64": {"threads": 128, "sass": "hgemm_nt_16x64", "params": "gemm", "share": "(16*64 + 32)*2 + (64*64 + 32)*2 + 4" },
    "hgemm_nn_32x64": {"threads": 128, "sass": "hgemm_nn_32x64", "params": "gemm", "share": "32*33*2 + 64*32*2 + 2048" },  #artificially limit occpancy
    "hgemm_nn_16x64": {"threads": 128, "sass": "hgemm_nn_16x64", "params": "gemm", "share": "(16*64 + 32)*2 + 64*64*2 + 4" },

    "sgemm_rnn_nn_128x32":    {"threads": 128, "sass": "sgemm_nn_rnn_128x32",       "params": "gemm_rnn",   "share": "(128*16 + 32)*2 + 32*16*2 + 4"},
    "sgemm_rnn_nn_vec_128x32":    {"threads": 128, "sass": "sgemm_nn_rnn_128x32",       "params": "gemm_rnn",   "share": "(128*16 + 32)*2 + 32*16*2 + 4", "args": {"vec": "1"}},

    "sgemm_rnn_bprop_tn_128x32":    {"threads": 128, "sass": "sgemm_tn_rnn_bprop_128x32",       "params": "gemm_rnn_bprop",   "share": "(128*16 + 32)*2 + 32*16*2 + 4"},
    "sgemm_rnn_bprop_tn_vec_128x32":    {"threads": 128, "sass": "sgemm_tn_rnn_bprop_128x32",       "params": "gemm_rnn_bprop",   "share": "(128*16 + 32)*2 + 32*16*2 + 4", "args": {"vec": "1"}},

    "persistent_rnn_fprop": {"threads": 256, "sass": "persistent_rnn_fprop", "params": "rnn_fprop", "share": "(64*48) + 4"},
    "persistent_rnn_bprop": {"threads": 256, "sass": "persistent_rnn_bprop", "params": "rnn_bprop", "share": "(64*48) + 4"},
}

_params = {
    "fprop": [
        "float* param_Sum",
        "float* param_X",
        "float* param_O",
        "float* param_I",
        "float* param_F",
        "float param_alpha",
        "float param_beta",
        "unsigned param_flags",
        "unsigned param_N",
        "unsigned param_K",
        "unsigned param_D",
        "unsigned param_H",
        "unsigned param_W",
        "unsigned param_WN",
        "unsigned param_HWN",
        "unsigned param_DHWN",
        "unsigned param_C",
        "unsigned param_KRST",
        "unsigned param_RST",
        "unsigned param_RS",
        "unsigned param_T",
        "unsigned param_R",
        "unsigned param_S",
        "unsigned param_magic_RS",
        "unsigned param_shift_RS",
        "unsigned param_magic_S",
        "unsigned param_shift_S",
        "int param_pad_d",
        "int param_pad_h",
        "int param_pad_w",
        "unsigned param_str_d",
        "unsigned param_str_h",
        "unsigned param_str_w",
        "unsigned param_dil_d",
        "unsigned param_dil_h",
        "unsigned param_dil_w",
        "unsigned param_P2",
        "unsigned param_Q",
        "unsigned param_PQk",
        "unsigned param_Qk",
        "unsigned param_k",
        "unsigned param_magic_PQk",
        "unsigned param_shift_PQk",
        "unsigned param_magic_Qk",
        "unsigned param_shift_Qk",
        "unsigned param_magic_k",
        "unsigned param_shift_k",
        "unsigned param_QN",
        "unsigned param_PQN",
        "unsigned param_MPQN",
        "unsigned param_gridN",
        "unsigned param_gridQN",
        "unsigned param_gridPQN",
        "unsigned param_gridMPQN",
    ],
    "fprop2": [
        "float* param_Sum",
        "float* param_X",
        "float* param_O",
        "float* param_I",
        "float* param_F",
        "float param_alpha",
        "float param_beta",
        "unsigned param_flags",
        "unsigned param_C",
        "unsigned param_D",
        "unsigned param_H",
        "unsigned param_W",
        "unsigned param_N",
        "unsigned param_K",
        "unsigned param_M",
        "unsigned param_P",
        "unsigned param_Q",
        "unsigned param_str_d",
        "unsigned param_str_h",
        "unsigned param_str_w",
        "int param_pad_d",
        "int param_pad_h",
        "int param_pad_w",
        "unsigned param_dil_d",
        "unsigned param_dil_h",
        "unsigned param_dil_w",
        "unsigned param_DHWN",
        "unsigned param_HWN",
        "unsigned param_WN",
        "unsigned param_MPQN",
        "unsigned param_PQN",
        "unsigned param_QN",
        "unsigned param_PQnk",
        "unsigned param_Qnk",
        "unsigned param_nk",
        "unsigned param_n",
        "unsigned param_k",
        "unsigned param_magic_PQnk",
        "unsigned param_shift_PQnk",
        "unsigned param_magic_Qnk",
        "unsigned param_shift_Qnk",
        "unsigned param_magic_nk",
        "unsigned param_shift_nk",
        "unsigned param_magic_k",
        "unsigned param_shift_k",
        "unsigned param_Km32",
        "unsigned param_K32p",
        "unsigned param_TRSK",
        "unsigned param_TRS",
        "unsigned param_RS",
        "unsigned param_S",
        "unsigned param_magic_RS",
        "unsigned param_shift_RS",
        "unsigned param_magic_S",
        "unsigned param_shift_S",
        "unsigned param_gridP2",
        "unsigned param_gridQ",
        "unsigned param_gridN",
        "unsigned param_gridQN",
        "unsigned param_gridPQN",
        "unsigned param_gridMPQN",
        "unsigned param_superM",
        "unsigned param_superP",
        "unsigned param_superQ",
        "unsigned param_superN",
        "unsigned param_shiftM",
        "unsigned param_shiftP",
        "unsigned param_shiftQ",
        "unsigned param_shiftN",
        "unsigned param_SuperM",
        "unsigned param_SuperP",
        "unsigned param_SuperQ",
        "unsigned param_SuperN",
    ],
    "updat2": [
        "float* param_F",
        "float* param_I",
        "float* param_E",
        "float param_alpha",
        "unsigned param_C",
        "unsigned param_D",
        "unsigned param_H",
        "unsigned param_W",
        "unsigned param_N",
        "unsigned param_K",
        "unsigned param_M",
        "unsigned param_P",
        "unsigned param_Q",
        "unsigned param_str_d",
        "unsigned param_str_h",
        "unsigned param_str_w",
        "int param_pad_d",
        "int param_pad_h",
        "int param_pad_w",
        "unsigned param_dil_d",
        "unsigned param_dil_h",
        "unsigned param_dil_w",
        "unsigned param_DHWN",
        "unsigned param_HWN",
        "unsigned param_WN",
        "unsigned param_MPQN16p",
        "unsigned param_MPQN",
        "unsigned param_PQN",
        "unsigned param_QN",
        "unsigned param_PQkc",
        "unsigned param_Qkc",
        "unsigned param_kc",
        "unsigned param_c",
        "unsigned param_k",
        "unsigned param_magic_PQkc",
        "unsigned param_shift_PQkc",
        "unsigned param_magic_Qkc",
        "unsigned param_shift_Qkc",
        "unsigned param_magic_kc",
        "unsigned param_shift_kc",
        "unsigned param_magic_c",
        "unsigned param_shift_c",
        "unsigned param_CRSTK",
        "unsigned param_CRST",
        "unsigned param_TRS",
        "unsigned param_RS",
        "unsigned param_S",
        "unsigned param_magic_TRS",
        "unsigned param_shift_TRS",
        "unsigned param_magic_RS",
        "unsigned param_shift_RS",
        "unsigned param_magic_S",
        "unsigned param_shift_S",
        "unsigned param_superM",
        "unsigned param_superP",
        "unsigned param_superQ",
        "unsigned param_superN",
        "unsigned param_shiftM",
        "unsigned param_shiftP",
        "unsigned param_shiftQ",
        "unsigned param_strideP",
        "unsigned param_strideQ",
        "unsigned param_stridePQ",
        "unsigned param_gridP",
        "unsigned param_gridQ",
        "unsigned param_loopX",
        "unsigned param_loopXp",
        "unsigned param_loopQ",
        "unsigned param_loopQp",
        "unsigned param_loopN",
        "unsigned param_loopNp",
    ],
    "gemm": [
        "float* param_C",
        "float* param_A",
        "float* param_B",
        "float param_alpha",
        "float param_beta",
        "unsigned param_flags",
        "unsigned param_lda",
        "unsigned param_ldb",
        "unsigned param_ldc",
        "unsigned param_m",
        "unsigned param_n",
        "unsigned param_k",
        "unsigned param_ldaz",
        "unsigned param_ldbz",
        "unsigned param_ldcz",
        "unsigned param_batch_loops",
    ],
    "gemm_rnn": [
        "float* param_C",
        "float* param_A",
        "float* param_B",
        "float* param_bias",
        "float* param_lock",
        "float param_alpha",
        "float param_beta",
        "float param_xcutoff",
        "int   param_flags",
        "int   param_lda",
        "int   param_ldb",
        "int   param_ldc",
        "int   param_m",
        "int   param_n",
        "int   param_k",
        "int   param_ldaz",
        "int   param_ldbz",
        "int   param_ldcz",
        "int   param_batch_loops",
        "int   param_dimB",
        "int   param_dimC",
        "int   param_unrolling",
        "int   param_numBlks",
        "int   param_numAblks"
    ],
    "gemm_rnn_bprop": [
        "float* param_C",
        "float* param_A",
        "float* param_B",
        "float* param_H",
        "float* param_lock",
        "float param_alpha",
        "float param_beta",
        "float param_xcutoff",
        "int   param_flags",
        "int   param_lda",
        "int   param_ldb",
        "int   param_ldc",
        "int   param_ldh",
        "int   param_m",
        "int   param_n",
        "int   param_k",
        "int   param_ldaz",
        "int   param_ldbz",
        "int   param_ldcz",
        "int   param_batch_loops",
        "int   param_dimB",
        "int   param_dimC",
        "int   param_dimH",
        "int   param_unrolling",
        "int   param_numBlks",
        "int   param_numAblks"
    ],
    "fpropw": [
        "float* param_S",
        "float* param_X",
        "float* param_O",
        "float* param_I",
        "float* param_F",
        "float param_alpha",
        "float param_beta",
        "unsigned param_flags",
        "unsigned param_C",
        "unsigned param_H",
        "unsigned param_P",
        "int param_pad_h",
        "int param_pad_w",
        "unsigned param_HWN",
        "unsigned param_WN",
        "unsigned param_PQN",
        "unsigned param_QN",
        "unsigned param_Qnk",
        "unsigned param_nk",
        "unsigned param_n",
        "unsigned param_k",
        "unsigned param_magic_Qnk",
        "unsigned param_shift_Qnk",
        "unsigned param_magic_nk",
        "unsigned param_shift_nk",
        "unsigned param_magic_k",
        "unsigned param_shift_k",
        "unsigned param_RSK",
        "unsigned param_4RSKp",
        "unsigned param_4HWNp",
        "unsigned param_gridK",
        "unsigned param_gridP2",
        "unsigned param_gridQ",
        "unsigned param_gridN",
        "unsigned param_gridQN",
        "unsigned param_gridPQN",
        "unsigned param_superP",
        "unsigned param_superQ",
        "unsigned param_superN",
        "unsigned param_shiftP",
        "unsigned param_shiftQ",
        "unsigned param_shiftN",
    ],
    "fpropw4X": [
        "float* param_S",
        "float* param_X",
        "float* param_O",
        "float* param_I",
        "float* param_F",
        "float param_alpha",
        "float param_beta",
        "unsigned param_flags",
        "unsigned param_C",
        "unsigned param_K",
        "unsigned param_N",
        "unsigned param_Xk",
        "unsigned param_k",
        "unsigned param_magic_Xk",
        "unsigned param_shift_Xk",
        "unsigned param_magic_k",
        "unsigned param_shift_k",
        "unsigned param_C_1152",
        "unsigned param_GXS_C_1152",
        "unsigned param_GYS_GXS_C_1152",
        "unsigned param_P",
        "unsigned param_Q",
        "unsigned param_QN",
        "unsigned param_PQN",
        "unsigned param_PQN15",
        "unsigned param_maskN",
        "unsigned param_shiftX",
        "unsigned param_shiftY",
        "unsigned param_superX",
        "unsigned param_superY",
        "unsigned param_gridN",
        "unsigned param_gridQN",
        "unsigned param_gridPQN",
    ],
    "fpropw4": [
        "float* param_S",
        "float* param_X",
        "float* param_O",
        "float* param_I",
        "float* param_F",
        "float param_alpha",
        "float param_beta",
        "unsigned param_flags",
        "unsigned param_C",
        "unsigned param_K",
        "unsigned param_N",
        "unsigned param_Y",
        "unsigned param_W",
        "unsigned param_YXN",
        "unsigned param_XN",
        "unsigned param_Y2",
        "unsigned param_GX",
        "unsigned param_Xk",
        "unsigned param_k",
        "unsigned param_magic_Xk",
        "unsigned param_shift_Xk",
        "unsigned param_magic_k",
        "unsigned param_shift_k",
        "unsigned param_P",
        "unsigned param_Q",
        "unsigned param_QN",
        "unsigned param_PQN",
        "unsigned param_PQN15",
        "unsigned param_maskN",
        "unsigned param_shiftX",
        "unsigned param_shiftY",
        "unsigned param_superX",
        "unsigned param_superY",
        "int param_pad_x",
        "int param_pad_y",
        "unsigned param_RSK",
        "unsigned param_RSK2p",
        "unsigned param_YXN2p",
        "unsigned param_gridN",
        "unsigned param_gridQN",
        "unsigned param_gridPQN",
    ],
    "fpropw5": [
        "float* param_O",
        "float* param_I",
        "float* param_F",
        "float param_alpha",
        "unsigned param_flags",
        "unsigned param_C",
        "unsigned param_K",
        "unsigned param_N",
        "unsigned param_H",
        "unsigned param_W",
        "unsigned param_HWN",
        "unsigned param_WN",
        "unsigned param_Y2",
        "unsigned param_GX",
        "unsigned param_Xk",
        "unsigned param_k",
        "unsigned param_magic_Xk",
        "unsigned param_shift_Xk",
        "unsigned param_magic_k",
        "unsigned param_shift_k",
        "unsigned param_P",
        "unsigned param_Q",
        "unsigned param_QN",
        "unsigned param_PQN",
        "unsigned param_PQNp",
        "unsigned param_PQN15p",
        "unsigned param_shiftY",
        "unsigned param_shiftX",
        "unsigned param_shiftN",
        "unsigned param_superY",
        "unsigned param_superX",
        "unsigned param_superN",
        "unsigned param_SuperY",
        "unsigned param_SuperX",
        "unsigned param_SuperN",
        "int param_pad_x",
        "int param_pad_y",
        "unsigned param_HWN2p",
        "unsigned param_C_1152",
    ],
    "updatw": [
        "float* param_F",
        "float* param_I",
        "float* param_E",
        "float param_alpha",
        "unsigned param_Y",
        "unsigned param_X",
        "unsigned param_P",
        "unsigned param_Q",
        "unsigned param_C",
        "unsigned param_K",
        "unsigned param_N",
        "int param_pad_y",
        "int param_pad_x",
        "unsigned param_GY",
        "unsigned param_GX",
        "unsigned param_GYS",
        "unsigned param_GXS",
        "unsigned param_shiftYI",
        "unsigned param_shiftXI",
        "unsigned param_superYI",
        "unsigned param_superXI",
        "unsigned param_superNI",
        "unsigned param_shiftY",
        "unsigned param_shiftX",
        "unsigned param_superY",
        "unsigned param_superX",
        "unsigned param_superN",
        "unsigned param_loopXI",
        "unsigned param_loopX",
        "unsigned param_loopN",
        "unsigned param_strideY",
        "unsigned param_strideX",
        "unsigned param_XN",
        "unsigned param_YXN",
        "unsigned param_QN",
        "unsigned param_PQN",
        "unsigned param_SK",
        "unsigned param_RSK",
        "unsigned param_Np",
        "unsigned param_XNp",
        "unsigned param_2XNp",
        "unsigned param_QNp",
        "unsigned param_CPQkc",
        "unsigned param_PQkc",
        "unsigned param_Qkc",
        "unsigned param_kc",
        "unsigned param_c",
        "unsigned param_k",
        "unsigned param_magic_CPQkc",
        "unsigned param_shift_CPQkc",
        "unsigned param_magic_PQkc",
        "unsigned param_shift_PQkc",
        "unsigned param_magic_Qkc",
        "unsigned param_shift_Qkc",
        "unsigned param_magic_kc",
        "unsigned param_shift_kc",
        "unsigned param_magic_c",
        "unsigned param_shift_c",
        "unsigned param_CRSK",
    ],
    "updatw4": [
        "float* param_F",
        "float* param_I",
        "float* param_E",
        "float param_alpha",
        "unsigned param_K",
        "unsigned param_C",
        "unsigned param_k",
        "unsigned param_c",
        "unsigned param_kc",
        "unsigned param_magic_kc",
        "unsigned param_shift_kc",
        "unsigned param_magic_c",
        "unsigned param_shift_c",
        "unsigned param_YXN2",
        "unsigned param_sYXN",
        "unsigned param_magic_sYXN",
        "unsigned param_shift_sYXN",
        "unsigned param_stride_YXNp",
        "unsigned param_YXN",
        "unsigned param_YXN_1152",
        "unsigned param_RSK",
        "unsigned param_CRSK",
        "unsigned param_Kp",
        "unsigned param_SKp",
        "unsigned param_RSK15_SK2p",
    ],
    "rnn_fprop": [
        "float* param_h",
        "float* param_hprev",
        "float* param_bias",
        "float* param_w",
        "int* param_lockAddr",
        "int param_ldh",
        "int param_ldw",
        "int param_bsz",
        "int param_seqLength",
        "int param_numBlks",
        "int param_rowSize",
        "int param_reverse",
        "float param_reluclip"
    ]
    ,
    "rnn_bprop": [
        "float* param_d",
        "float* param_dnext",
        "float* param_h",
        "float* param_w",
        "int* param_lockAddr",
        "int param_ldd",
        "int param_ldh",
        "int param_ldw",
        "int param_bsz",
        "int param_seqLength",
        "int param_numBlks",
        "int param_rowSize",
        "int param_reverse",
        "float param_reluclip"
    ]
}

_params["bprop"] = _params["fprop"] + [
        "unsigned param_magic_str_d",
        "unsigned param_shift_str_d",
        "unsigned param_magic_str_h",
        "unsigned param_shift_str_h",
        "unsigned param_magic_str_w",
        "unsigned param_shift_str_w",
    ]
_params["bprop2"] = _params["fprop2"] + [
        "unsigned param_magic_str_d",
        "unsigned param_shift_str_d",
        "unsigned param_magic_str_h",
        "unsigned param_shift_str_h",
        "unsigned param_magic_str_w",
        "unsigned param_shift_str_w",
    ]

_space_re = re.compile(r"\s+")

_share_template = r"""
    .shared .align 4 .b32 share[{0}];
"""

_kernel_template = r"""
.version {6}
.target {0}
.address_size 64

// args: {5}

.visible .entry  {1}(
{2}
)
.reqntid {3}
{{
{4}
    ret;
}}
"""

def get_ptx_file(kernel_spec, kernel_name, arch, ptx_ver):

    thread_spec = kernel_spec["threads"]
    args_spec   = str(kernel_spec.get("args",""))
    param_spec  = _params[kernel_spec["params"]]

    kernel_params = []
    for p in param_spec:
        ptype, pname = _space_re.split(p)

        if ptype[-1] == '*':
            ptype = '.u64'
        elif ptype == 'float':
            ptype = '.f32'
        else:
            ptype = '.u32'

        kernel_params.append("    .param %s %s" % (ptype, pname))

    kernel_params = ",\n".join(kernel_params)

    if "share" in kernel_spec:
        share = _share_template.format(eval(kernel_spec["share"]))
    else:
        share = ""

    kernel_text = _kernel_template.format(arch, kernel_name, kernel_params, thread_spec, share, args_spec, ptx_ver)
    kernel_ptx  = os.path.join(ptx_dir, kernel_name + ".ptx")

    current_text = ""
    if os.path.exists(kernel_ptx):
        f = open(kernel_ptx, "r")
        current_text = f.read()
        f.close()
    # only write out the kernel if text has changed.
    if kernel_text != current_text:
        f = open(kernel_ptx, "w")
        f.write(kernel_text)
        f.close()

    return kernel_ptx


include_re = re.compile(r'^<INCLUDE\s+file="([^"]+)"\s*/>')

def extract_includes(name, includes=None):
    if not includes:
        includes = list()
    sass_file = os.path.join(sass_dir, name)
    includes.append((sass_file, os.path.getmtime(sass_file)))
    for line in open(sass_file, "r"):
        match = include_re.search(line)
        if match:
            extract_includes(match.group(1), includes)
    return includes

def run_command(cmdlist):
    cmd  = " ".join(cmdlist)
    proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = proc.communicate()
    if proc.returncode:
        raise RuntimeError("Error(%d):\n%s\n%s" % (proc.returncode, cmd, err))
    if debug:
        neon_logger.display(cmd)
        if out: neon_logger.display(out)
        if err: neon_logger.display(err)

@context_dependent_memoize
def get_kernel(base_name, options=None):

    attributes = drv.Context.get_device().get_attributes()
    major = attributes[drv.device_attribute.COMPUTE_CAPABILITY_MAJOR]
    minor = attributes[drv.device_attribute.COMPUTE_CAPABILITY_MINOR]
    if major < 5:
        raise RuntimeError("sass kernels require Maxwell or greater class hardware")

    arch = "sm_%d%d" % (major, minor)

    libprefix = "PERL5LIB=%s" % (maxas_dir)
    maxas_i = [libprefix, os.path.join(maxas_dir, "maxas.pl") + " -i -w"]
    maxas_p = [libprefix, os.path.join(maxas_dir, "maxas.pl") + " -p"]

    kernel_spec = kernels[base_name]
    kernel_name = base_name

    if "args" in kernel_spec:
        for pair in kernel_spec["args"].items():
            maxas_i.append("-D%s %s" % pair)
            maxas_p.append("-D%s %s" % pair)

    if options is not None:
        for opt in options:
            if type(opt) is tuple:
                maxas_i.append("-D%s %s" % opt)
                maxas_p.append("-D%s %s" % opt)
                kernel_name += "_%s%s" % opt
            else:
                maxas_i.append("-D%s 1" % opt)
                maxas_p.append("-D%s 1" % opt)
                kernel_name += "_%s" % opt

    maxas_i.insert(2, "-k " + kernel_name)

    sass_name  = kernel_spec["sass"] + ".sass"
    cubin_name = kernel_name + ".cubin"

    ptx_version = "4.2" if major < 6 else "5.0"
    ptx_file   = get_ptx_file(kernel_spec, kernel_name, arch, ptx_version)
    sass_file  = os.path.join(sass_dir, sass_name)
    cubin_file = os.path.join(cubin_dir, cubin_name)

    if not os.path.exists(sass_file):
        raise RuntimeError("Missing: %s for kernel: %s" % (sass_name, kernel_name))

    ptx_age   = os.path.getmtime(ptx_file)
    cubin_age = os.path.getmtime(cubin_file) if os.path.exists(cubin_file) else 0

    if ptx_age > cubin_age:
        run_command([ "ptxas -v -arch", arch, "-o", cubin_file, ptx_file ])
        cubin_age = 0

    includes = extract_includes(sass_name)

    for include, include_age in includes:
        if include_age > cubin_age:
            run_command(maxas_i + [sass_file, cubin_file])
            cubin_age = include_age
            break

    if debug:
        pre_file  = os.path.join(pre_dir,  kernel_name + "_pre.sass")
        dump_file = os.path.join(dump_dir, kernel_name + "_dump.sass")
        pre_age   = os.path.getmtime(pre_file)  if os.path.exists(pre_file)  else 0
        dump_age  = os.path.getmtime(dump_file) if os.path.exists(dump_file) else 0

        for include, include_age in includes:
            if include_age > pre_age:
                run_command(maxas_p + [sass_file, pre_file])
                break

        if cubin_age > dump_age:
            run_command(["nvdisasm -raw", cubin_file, ">", dump_file])

    params  = _params[kernel_spec["params"]]
    sig = ""
    for p in params:
        ptype, pname = _space_re.split(p)
        if ptype[-1] == '*':
            sig += "Q"
        elif ptype == 'float':
            sig += "f"
        elif ptype == 'unsigned':
            sig += "I"
        else:
            sig += "i"

    module = drv.module_from_file(os.path.join(cubin_dir, kernel_name + ".cubin"))
    func   = module.get_function(kernel_name)
    func.prepare(sig)
    func.threads = kernel_spec["threads"]
    return func





# ----------------------------------------------------------------------------
# Copyright 2014-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
from __future__ import division
from neon.transforms.transform import Transform


class Identity(Transform):
    """
    Identity activation function, :math:`f(x) = x`
    """
    def __init__(self, name=None):
        """
        Class constructor.
        """
        super(Identity, self).__init__(name)

    def __call__(self, x):
        """
        Returns the input as output.

        Arguments:
            x (Tensor or optree): input value

        Returns:
            Tensor or optree: identical to input
        """
        return x

    def bprop(self, x):
        """
        Returns the derivative.

        Arguments:
            x (Tensor or optree): Input value

        Returns:
            Integer value 1.

        """
        return 1


class Rectlin(Transform):
    """
    Rectified Linear Unit (ReLu) activation function, :math:`f(x) = \max(x, 0)`.
    Can optionally set a slope which will make this a Leaky ReLu.
    """
    def __init__(self, slope=0, name=None):
        """
        Class constructor.

        Args:
            slope (float, optional): Slope for negative domain. Defaults to 0.
            name (string, optional): Name to assign this class instance.
        """
        super(Rectlin, self).__init__(name)
        self.slope = slope
        self.is_mklop = True

    def __call__(self, x, nglayer=None):
        """
        Returns the Exponential Linear activation

        Arguments:
            x (Tensor or optree): Input value

        Returns:
            Tensor or optree: output activation
        """
        return self.be.fprop_relu(nglayer, x, self.slope)

    def bprop(self, x, nglayer=None, error=None, deltas=None):
        """
        Returns the derivative.

        Arguments:
            x (Tensor or optree): Input value

        Returns:
            Tensor or optree: Derivative
        """
        return self.be.bprop_relu(nglayer, x, error, deltas, self.slope)


class Rectlinclip(Transform):
    """
    Clipped ReLu activation function
    Computes the function f(x) = min(max(0, x),cutoff)
    """
    def __init__(self, slope=0, name=None, xcut=20.0):
        super(Rectlinclip, self).__init__(name)
        self.xcut = xcut
        self.slope = slope

    def __call__(self, x):
        return self.be.minimum(self.be.maximum(x, 0) + self.slope * self.be.minimum(x, 0),
                               self.xcut)

    def bprop(self, x):
        return (self.be.greater(x, 0) + self.slope * self.be.less(x, 0)) *\
                self.be.greater(self.xcut, x)


class Explin(Transform):
    """
    Exponential Linear activation function, :math:`f(x) = \max(x, 0) + \\alpha (e^{\min(x, 0)}-1)`

    From: Clevert, Unterthiner and Hochreiter, ICLR 2016.
    """
    def __init__(self, alpha=1.0, name=None):
        """
        Class constructor.

        Arguments:
            alpha (float): weight of exponential factor for negative values (default: 1.0).
            name (string, optional): Name (default: None)
        """
        super(Explin, self).__init__(name)
        self.alpha = alpha

    def __call__(self, x):
        """
        Returns the Exponential Linear activation

        Arguments:
            x (Tensor or optree): input value

        Returns:
            Tensor or optree: output activation
        """
        return self.be.maximum(x, 0) + self.alpha * (self.be.exp(self.be.minimum(x, 0)) - 1)

    def bprop(self, x):
        """
        Returns the derivative.

        Arguments:
            x (Tensor or optree): Input value

        Returns:
            Tensor or optree: Derivative
        """
        return self.be.greater(x, 0) + self.be.minimum(x, 0) + self.alpha * self.be.less(x, 0)


class Normalizer(Transform):
    """
    Normalize inputs by a fixed divisor.
    """
    def __init__(self, name=None, divisor=128.):
        """
        Class constructor.

        Arguments:
            divisor (float, optional): Normalization factor (default: 128)
            name (string, optional): Name (default: None)
        """
        super(Normalizer, self).__init__(name)
        self.divisor = divisor

    def __call__(self, x):
        """
        Returns the normalized value.

        Arguments:
            x (Tensor or optree): Input value

        Returns:
            Tensor or optree: Output :math:`x / N`
        """
        return x / self.divisor

    def bprop(self, x):
        """
        Returns the derivative.

        Arguments:
            x (Tensor or optree): Input value

        Returns:
            Tensor or optree: Derivative
        """
        return 1.0 / self.divisor


class Softmax(Transform):
    """
    SoftMax activation function. Ensures that the activation output sums to 1.
    """
    def __init__(self, axis=0, name=None, epsilon=2**-23):
        """
        Class constructor.

        Arguments:
            name (string, optional): Name (default: none)
            epsilon (float, optional): Not used.
            axis (int, optional): axis to perform softmax (default: 0)
        """
        super(Softmax, self).__init__(name)
        self.epsilon = epsilon
        self.axis = axis

    def __call__(self, x):
        """
        Returns the Softmax value.

        Arguments:
            x (Tensor or optree): Input value

        Returns:
            Tensor or optree: Output activation
        """
        return self.be.fprop_softmax(x, self.axis)

    def bprop(self, x):
        """
        Returns the derivative.

        Arguments:
            x (Tensor or optree): Input value

        Returns:
            Integer value 1
        """
        return 1


class PixelwiseSoftmax(Transform):
    """
    Pixelwise SoftMax activation function.
    Computes the function f(x_k) = exp(x_k) / sum_i(exp(x_i))
    """
    def __init__(self, c, name=None, epsilon=2**-23):
        super(PixelwiseSoftmax, self).__init__(name)
        self.epsilon = epsilon
        self.c = c

    def __call__(self, x):
        y = x.reshape((self.c, -1))
        y[:] = (self.be.reciprocal(self.be.sum(self.be.exp(y - self.be.max(y, axis=0)), axis=0)) *
                self.be.exp(y - self.be.max(y, axis=0)))
        return x

    def bprop(self, x):
        return 1


class Tanh(Transform):
    """
    Hyperbolic tangent activation function, :math:`f(x) = \\tanh(x)`.
    """
    def __init__(self, name=None):
        """
        Class constructor.
        """
        super(Tanh, self).__init__(name)

    def __call__(self, x):
        """
        Returns the hyperbolic tangent.

        Arguments:
            x (Tensor or optree): Input value

        Returns:
            Tensor or optree: Output activation
        """
        return self.be.tanh(x)

    def bprop(self, x):
        """
        Returns the derivative.

        Arguments:
            x (Tensor or optree): Input value

        Returns:
            Tensor or optree: Derivative, :math:`1-x^2`
        """
        return (1.0 - self.be.square(x))


class Logistic(Transform):
    """
    Logistic sigmoid activation function, :math:`f(x) = 1 / (1 + \exp(-x))`

    Squashes the input from range :math:`[-\infty,+\infty]` to :math:`[0, 1]`
    """
    def __init__(self, name=None, shortcut=False):
        """
        Initialize Logistic based on whether shortcut is True or False. Shortcut
        should be set to true when Logistic is used in conjunction with a CrossEntropy cost.
        Doing so allows a shortcut calculation to be used during backpropagation.

        Args:
            shortcut (bool): If True, shortcut calculation will be used during backpropagation.

        """
        super(Logistic, self).__init__(name=name)

        self.set_shortcut(shortcut)

    def set_shortcut(self, shortcut):
        """
        Sets the backpropagation to use the shortcut when gradients do not
        need to be calculated.

        If True, a shortcut calculation is used. If False, the actual derivative
        is return during backpropagation.

        Arguments:
            shortcut (bool): If True, shortcut calculation will be used during backpropagation.
        """
        self.shortcut = shortcut

        if shortcut:
            self.bprop_func = lambda x: 1
        else:
            self.bprop_func = lambda x: x * (1.0 - x)

    def __call__(self, x):
        """
        Returns the sigmoidal activation.

        Arguments:
            x (Tensor or optree): Input value

        Returns:
            Tensor or optree: Output activation
        """
        return self.be.sig(x)

    def bprop(self, y):
        """

        Returns the derivative of the logistic (sigmoid) function at y (output)

        Args:
            y (Tensor or OpTree): input. y = f(x)

        Returns:
            OpTree: Derivative of the Logistic (sigmoid)
                    Returns 1 if shortcut is True.
                    Returns derivative (y*(1-y)) if shortcut is False.

        """
        return self.bprop_func(y)


class Sign(Transform):
    """
    Sign activation function.
    Computes the function f(x) = Sign(x).
    Uses straight-through estimator for bprop.
    """
    def __init__(self, name=None):
        super(Sign, self).__init__(name)

    def __call__(self, x):
        self.inputs = self.be.array(x.get())
        return self.be.binarize(x, x, stochastic=False)

    def bprop(self, x):
        return self.be.less_equal(self.be.absolute(self.inputs), 1)



import numpy as np
from neon import NervanaObject
import logging


logger = logging.getLogger(__name__)


class BeamSearch(NervanaObject):
    """
    Beam search for Encoder-Decoder models.

    Arguments:
        seq2seq (Object): Seq2Seq container object with a trained model to use for inference
    """
    def __init__(self, seq2seq):
        super(BeamSearch, self).__init__(name=None)

        self.layers = seq2seq

        self.hasLUT = seq2seq.hasLUT

        # resize buffers
        new_steps = 1
        seq2seq.decoder.switch_mode(inference=True)
        self.z_shape = new_steps if self.hasLUT else (seq2seq.out_shape[0], new_steps)

    def beamsearch(self, inputs, num_beams=5, steps=None):
        """
        Perform an fprop path and beam search on a given set of network inputs.

        Arguments:
            inputs (Tensor): Minibatch of network inputs
            num_beams (Int): Number of beams (hypothesis) to search over
            steps (Int): Length of desired output in number of time steps
        """
        self.num_beams = num_beams

        self.num_live = 1
        self.num_dead = 0

        # allocate space for outputs, hypotheses and associated scores
        # default to the number of steps used by this decoder most recently.
        if steps is None:
            steps = self.layers.in_shape[1]

        bsz = self.be.bsz

        self.z_list = [self.be.iobuf(self.z_shape) for _ in range(num_beams)]
        for i in range(num_beams):
            if getattr(self.layers.decoder, 'start_index', None) is not None:
                self.z_list[i][self.layers.decoder.start_index] = 1

        self.candidates = [np.zeros((steps, bsz)) for _ in range(num_beams)]
        self.scores = [np.zeros(bsz) for _ in range(num_beams)]

        z = self.be.iobuf(self.z_shape)

        # encoder
        self.layers.encoder.fprop(inputs, inference=True, beta=0.0)
        final_states = self.layers.encoder.get_final_states(
            self.layers.decoder_connections
        )

        if len(final_states) != len(self.layers.decoder._recurrent):
            raise ValueError((
                'number of decoder layers ({num_layers}) does not match '
                'the number of decoder connections ({num_decoder_connections}).'
            ).format(
                num_layers=len(self.layers.decoder._recurrent),
                num_decoder_connections=len(final_states),
            ))
        else:
            init_state_list = final_states

        # num_beams x num_layers x layer_shape
        self.init_state_lists = [[self.be.zeros_like(rec.h[-1])
                                 for rec in self.layers.decoder._recurrent]
                                 for _ in range(num_beams)]  # TODO: rename
        self.init_state_lists[0] = init_state_list

        # decoder
        if self.hasLUT:
            z_beams = [self.be.iobuf(self.layers.decoder.out_shape) for _ in range(self.num_beams)]
        else:
            z_beams = [self.be.iobuf(self.z_shape) for _ in range(self.num_beams)]

        # num_beams x num_layers x layer_shape
        hidden_state_beams = [[self.be.zeros_like(l.final_state())
                               for l in self.layers.decoder._recurrent]
                              for _ in range(self.num_beams)]  # will be list of lists

        for t in range(steps):
            # fprop all beams
            for bb in range(self.num_live):
                z = self.layers.decoder.fprop(self.z_list[bb], inference=True,
                                              init_state_list=self.init_state_lists[bb])
                z_beams[bb][:] = z
                for i, l in enumerate(self.layers.decoder._recurrent):
                    hidden_state_beams[bb][i][:] = l.final_state()
            # search over all beams
            self._beamsearch_step(t, z_beams, hidden_state_beams)

        return self.be.array(self.candidates[-1])

    def _beamsearch_step(self, t, z_list_t, init_state_lists):
        """
        Arguments:
            t (int): time step
            z_list (list of tensors) : fprop outputs for all beams
        """
        num_out = self.layers.out_shape[0]

        # bring init_state_lists list of lists over to CPU
        # init_state_lists_in, init_state_lists_out
        init_state_lists_in = []
        for isl in init_state_lists:
            init_state_lists_in.append([state.get() for state in isl])

        # num_beams x num_layers x layer_hidden_shape
        init_state_lists_out = [[np.zeros(tnsr.shape) for tnsr in init_state_lists[0]]
                                for _ in range(len(init_state_lists_in))]

        # pick highest scoring across all beams
        z_list = [z_list_t[bb].get() for bb in range(self.num_live)]
        scores_list = [np.log(z) + score for z, score in zip(z_list, self.scores)]
        scores = np.vstack(scores_list)
        self.num_live = num_live = self.num_beams - self.num_dead
        ind = scores.argsort(axis=0)[-num_live:]  # top scores across both old beams and new tokens
        self.scores_list = scores_list  # for testing

        # store candidates and scores for all beams and samples in batch
        hyp_num = (ind // num_out).astype(np.int32)  # hypothesis number
        word_num = ind % num_out  # onehot word index

        # create a copy of candidates buffer to read from so it does not get overwritten
        old_candidates = [can.copy() for can in self.candidates]

        for bb in range(num_live):
            for hyp in range(len(scores_list)):
                ex_with_hyp = np.where(hyp_num[bb, :] == hyp)[0]
                self.candidates[bb][:, ex_with_hyp] = old_candidates[hyp][:, ex_with_hyp]
                wnum = word_num[bb, ex_with_hyp]
                self.scores[bb][ex_with_hyp] = scores_list[hyp][wnum, ex_with_hyp]
            self.candidates[bb][t, :] = word_num[bb, :]

        # get decoder hidden states for each beam
        for bb in range(num_live):
            for hyp in range(len(scores_list)):
                ex_with_hyp = np.where(hyp_num[bb, :] == hyp)[0]

                # transfer hidden state from DECODER to next step
                for ii in range(len(self.layers.decoder._recurrent)):
                    init_state_lists_out[bb][ii][:, ex_with_hyp] = \
                        init_state_lists_in[hyp][ii][:, ex_with_hyp]

        # get inputs for next step for each beam
        for bb in range(num_live):
            # create input for next time step
            if self.hasLUT:
                # use index. TODO: Backend needs to catch these striding mismatches
                inputs = self.be.array(self.candidates[bb][t, :].reshape(1, -1))
                self.z_list[bb][:] = inputs
            else:
                # convoluted softmax -> index -> onehot scheme
                inputs = self.be.array(self.candidates[bb][t, :], dtype=np.int32)
                self.z_list[bb][:] = self.be.onehot(inputs, axis=0)

        # copy list of lists of numpy matrices into list of lists of GPU tensors
        for ii, isl in enumerate(init_state_lists_out):
            for jj, state in enumerate(isl):
                self.init_state_lists[ii][jj][:] = state



#!/usr/bin/python

# Copyright 2015-2016 Nervana Systems Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import division
import numpy as np
from struct import pack, unpack
from neon import logger as neon_logger

def ceil_div(x, y):
    return -(-x // y)

def out_dim(S, X, padding, strides):
    return ceil_div(X - S + 1 + 2 * padding, strides)

def strip_mantissa(val):
    i = unpack('I', pack('f', val))[0] & 0x7f800000
    f = unpack('f', pack('I', i))[0]
    return f

def quantize(ary, bits):
    maxval = float(np.max(np.absolute(ary)))
    scale  = strip_mantissa(maxval) / float(1 << bits - 2)
    ary    = np.around(ary * (1.0 / scale)).astype(np.int64)
    return ary, np.float32(scale)

######### Direct Convolution #########

def fconv_slice(q, S, X, padding, strides):
    f1 = 0
    f2 = S - 1
    x1 = q * strides - padding
    x2 = x1 + f2
    if x1 < 0:
        f1 = -x1
        x1 = 0
    if x2 >= X:
        dif = x2 - X + 1
        f2 -= dif
        x2 -= dif
    return (slice(f1, f2 + 1), slice(x1, x2 + 1), f2 - f1 + 1)

def bconv_slice(x, S, Q, padding, strides):
    qs = x - (S - padding - 1)
    firstF = None
    for s in range(S): #TODO remove loop logic here.
        q = qs + s
        if q % strides == 0:
            q //= strides
            if q >= 0 and q < Q:
                if firstF is None:
                    firstF = s
                    firstE = q
                lastF = s
                lastE = q
    if firstF is None:
        return (slice(0,0,1), slice(0,0,1), 0)
    return (slice(firstF,lastF + 1,strides), slice(firstE,lastE + 1,1), 0)

def xprop_direct(I, F, O, padding, strides, backward=False):

    if all(x == 1 for x in F.shape[1:3]):
        C = F.shape[0]
        if backward:
            # CxHWN = CxK . KxHWN
            O[:] = np.dot( F.reshape((C, -1)),   I.reshape((C, -1)) ).reshape((O.shape))
        else:
            # KxHWN = CxK.T . CxHWN
            O[:] = np.dot( F.reshape((C, -1)).T, I.reshape((C, -1)) ).reshape((O.shape))
        return

    if backward:
        # C <=> K and mirror R,S
        F = np.transpose(F[:,::-1,::-1,:], (3,1,2,0)).copy()
        xconv_slice = bconv_slice
    else:
        xconv_slice = fconv_slice

    C, Y, X, N = I.shape
    C, R, S, K = F.shape
    K, P, Q, N = O.shape

    qSlice = [ xconv_slice(q, S, X, padding[0], strides[0]) for q in range(Q) ]

    for p in range(P):
        sliceR, sliceY, _ = xconv_slice(p, R, Y, padding[1], strides[1])

        for q in range(Q):
            sliceS, sliceX, _ = qSlice[q]

            slicedF = F[:,sliceR,sliceS,:].reshape((-1, K))
            slicedI = I[:,sliceY,sliceX,:].reshape((-1, N))

            O[:,p,q,:] = np.dot( slicedF.T,  slicedI )

def updat_direct(I, E, U, padding, strides):

    C, Y, X, N = I.shape
    K, P, Q, N = E.shape
    C, R, S, K = U.shape

    if all(x == 1 for x in (R, S)):
        # CxK = CxHWN . KxHWN.T
        U[:] = np.dot( I.reshape((C, -1)), E.reshape((K, -1)).T ).reshape((U.shape))
        return

    U.fill(0.0)

    qSlice = [ fconv_slice(q, S, X, padding[0], strides[0]) for q in range(Q) ]

    for p in range(P):
        sliceR, sliceY, rlen = fconv_slice(p, R, Y, padding[1], strides[1])

        for q in range(Q):
            sliceS, sliceX, slen = qSlice[q]

            slicedI = I[:,sliceY,sliceX,:].reshape((-1, N))
            slicedE = E[:,p,q,:]

            U[:,sliceR,sliceS,:] += np.dot( slicedI,  slicedE.T ).reshape((C, rlen, slen, K))

######### Winograd Convolution #########

I_2x2_3x3 = np.array([
    [ 1.0,  0.0, -1.0,  0.0 ],
    [ 0.0,  1.0,  1.0,  0.0 ],
    [ 0.0, -1.0,  1.0,  0.0 ],
    [ 0.0,  1.0,  0.0, -1.0 ]])
F_2x2_3x3 = np.array([
    [ 1.0,  0.0, 0.0 ],
    [ 0.5,  0.5, 0.5 ],
    [ 0.5, -0.5, 0.5 ],
    [ 0.0,  0.0, 1.0 ]])
O_2x2_3x3 = np.array([
    [ 1.0, 1.0,  1.0,  0.0 ],
    [ 0.0, 1.0, -1.0, -1.0 ]]) #, dtype=np.float32

half    = np.float(0.5)
quarter = np.float(0.25)

def trans_I_2x2_3x3(Iw, I, minimal=False):
    if minimal:

        T0 = np.empty((4,4))
        T1 = np.empty((4,4))

        for O, I in ((T0, I), (T1, T0.T)):

            O[0,:] = I[0,:] - I[2,:]
            O[1,:] = I[1,:] + I[2,:]
            O[2,:] = I[2,:] - I[1,:]
            O[3,:] = I[1,:] - I[3,:]

        Iw[:] = T1.T

        # TI00 = I[0,0] - I[2,0]
        # TI01 = I[0,1] - I[2,1]
        # TI02 = I[0,2] - I[2,2]
        # TI03 = I[0,3] - I[2,3]
        # TI30 = I[1,0] - I[3,0]
        # TI31 = I[1,1] - I[3,1]
        # TI32 = I[1,2] - I[3,2]
        # TI33 = I[1,3] - I[3,3]
        # TI10 = I[1,0] + I[2,0]
        # TI11 = I[1,1] + I[2,1]
        # TI12 = I[1,2] + I[2,2]
        # TI13 = I[1,3] + I[2,3]
        # TI20 = I[2,0] - I[1,0]
        # TI21 = I[2,1] - I[1,1]
        # TI22 = I[2,2] - I[1,2]
        # TI23 = I[2,3] - I[1,3]

        # Iw[0,0] = TI00 - TI02
        # Iw[0,3] = TI01 - TI03
        # Iw[3,0] = TI30 - TI32
        # Iw[3,3] = TI31 - TI33
        # Iw[1,0] = TI10 - TI12
        # Iw[2,0] = TI20 - TI22
        # Iw[1,3] = TI11 - TI13
        # Iw[2,3] = TI21 - TI23
        # Iw[2,1] = TI21 + TI22
        # Iw[2,2] = TI22 - TI21
        # Iw[0,1] = TI01 + TI02
        # Iw[0,2] = TI02 - TI01
        # Iw[1,1] = TI11 + TI12
        # Iw[1,2] = TI12 - TI11
        # Iw[3,1] = TI31 + TI32
        # Iw[3,2] = TI32 - TI31

    else:
        Iw[:] = np.dot( np.dot(I_2x2_3x3, I), I_2x2_3x3.T )

def trans_F_2x2_3x3(Fw, F, minimal=False):
    if minimal:

        T0 = np.empty((4,3))
        T1 = np.empty((4,4))

        for O, I in ((T0, F), (T1, T0.T)):

            t0 = (I[0,:] + I[2,:]) * 0.5

            O[0,:] = I[0,:]
            O[1,:] = t0 + I[1,:] * 0.5
            O[2,:] = t0 - I[1,:] * 0.5
            O[3,:] = I[2,:]

        Fw[:] = T1.T

        # TF00 = F[0,0]
        # TF01 = F[0,1]
        # TF02 = F[0,2]
        # TF30 = F[2,0]
        # TF31 = F[2,1]
        # TF32 = F[2,2]
        # Fw[0,0] = TF00
        # Fw[0,3] = TF02
        # Fw[3,0] = TF30
        # Fw[3,3] = TF32
        # tb0 = TF00 + TF02
        # tb3 = TF30 + TF32
        # ta0 = F[0,0] + F[2,0]
        # ta1 = F[0,1] + F[2,1]
        # ta2 = F[0,2] + F[2,2]
        # tb0 = tb0 * 0.5
        # tb3 = tb3 * 0.5
        # ta0 = ta0 * 0.5
        # ta1 = ta1 * 0.5
        # ta2 = ta2 * 0.5
        # Fw[0,1] = tb0 + TF01*0.5
        # Fw[0,2] = tb0 - TF01*0.5
        # Fw[3,1] = tb3 + TF31*0.5
        # Fw[3,2] = tb3 - TF31*0.5
        # TF10 = ta0 + F[1,0]*0.5
        # TF20 = ta0 - F[1,0]*0.5
        # TF11 = ta1 + F[1,1]*0.5
        # TF21 = ta1 - F[1,1]*0.5
        # TF12 = ta2 + F[1,2]*0.5
        # TF22 = ta2 - F[1,2]*0.5
        # Fw[1,0] = TF10
        # Fw[2,0] = TF20
        # Fw[1,3] = TF12
        # Fw[2,3] = TF22
        # tb1 = TF10 + TF12
        # tb2 = TF20 + TF22
        # tb1 = tb1 * 0.5
        # tb2 = tb2 * 0.5
        # Fw[1,1] = tb1 + TF11*0.5
        # Fw[1,2] = tb1 - TF11*0.5
        # Fw[2,1] = tb2 + TF21*0.5
        # Fw[2,2] = tb2 - TF21*0.5

    else:
        Fw[:] = np.dot( np.dot(F_2x2_3x3, F), F_2x2_3x3.T )

def trans_O_2x2_3x3(Mw, minimal=False):
    if minimal:

        T0 = np.empty((2,4))
        T1 = np.empty((2,2))

        for O, I in ((T0, Mw), (T1, T0.T)):

            t0 = I[0,:] + I[1,:]
            t1 = I[1,:] - I[2,:]

            O[0,:] = t0 + I[2,:]
            O[1,:] = t1 - I[3,:]

        return T1.T

    else:
        return np.dot( np.dot(O_2x2_3x3, Mw), O_2x2_3x3.T )


I_3x3_2x2 = np.array([
    [ 1.0,  0.0, -1.0,  0.0 ],
    [ 0.0,  1.0,  1.0,  0.0 ],
    [ 0.0, -1.0,  1.0,  0.0 ],
    [ 0.0, -1.0,  0.0,  1.0 ]])
F_3x3_2x2 = np.array([
    [ 1.0,  0.0 ],
    [ 0.5,  0.5 ],
    [ 0.5, -0.5 ],
    [ 0.0,  1.0 ]])
O_3x3_2x2 = np.array([
    [ 1.0, 1.0,  1.0,  0.0 ],
    [ 0.0, 1.0, -1.0,  0.0 ],
    [ 0.0, 1.0,  1.0,  1.0 ]])

def trans_I_3x3_2x2(Iw, I, minimal=False):
    if minimal:

        T0 = np.empty((4,4))
        T1 = np.empty((4,4))

        for O, I in ((T0, I), (T1, T0.T)):

            O[0,:] = I[0,:] - I[2,:]
            O[1,:] = I[1,:] + I[2,:]
            O[2,:] = I[2,:] - I[1,:]
            O[3,:] = I[3,:] - I[1,:]

        Iw[:] = T1.T

    else:
        Iw[:] = np.dot( np.dot(I_3x3_2x2, I), I_3x3_2x2.T )

def trans_F_3x3_2x2(Fw, F, minimal=False):
    if minimal:

        T0 = np.empty((4,2))
        T1 = np.empty((4,4))

        for O, I in ((T0, F), (T1, T0.T)):

            O[0,:] =  I[0,:]
            O[1,:] = (I[0,:] + I[1,:]) * 0.5
            O[2,:] = (I[0,:] - I[1,:]) * 0.5
            O[3,:] =  I[1,:]

        Fw[:] = T1.T

        # x0  =  half*F[0,0]
        # x1  =  half*F[1,1]
        # B0  =  half*F[1,0] + x0
        # B1  =  half*F[0,1] + x1
        # C0  = -half*F[1,0] + x0
        # C1  =  half*F[0,1] - x1
        # x2  =  half*B0
        # x3  =  half*C0
        # Fw[0,0] =  F[0,0]
        # Fw[0,1] =  half*F[0,1] + x0
        # Fw[0,2] = -half*F[0,1] + x0
        # Fw[0,3] =  F[0,1]
        # Fw[3,0] =  F[1,0]
        # Fw[3,1] =  half*F[1,0] + x1
        # Fw[3,2] =  half*F[1,0] - x1
        # Fw[3,3] =  F[1,1]
        # Fw[1,0] =  B0
        # Fw[1,1] =  half*B1 + x2
        # Fw[1,2] = -half*B1 + x2
        # Fw[1,3] =  B1
        # Fw[2,0] =  C0
        # Fw[2,1] =  half*C1 + x3
        # Fw[2,2] = -half*C1 + x3
        # Fw[2,3] =  C1
    else:
        Fw[:] = np.dot( np.dot(F_3x3_2x2, F), F_3x3_2x2.T )

def trans_O_3x3_2x2(Mw, minimal=False):

    if minimal:

        T0 = np.empty((3,4))
        T1 = np.empty((3,3))

        for O, I in ((T0, Mw), (T1, T0.T)):

            t0 = I[1,:] + I[2,:]

            O[0,:] = t0 + I[0,:]
            O[1,:] = I[1,:] - I[2,:]
            O[2,:] = t0 + I[3,:]

        return T1.T
    else:
        return np.dot( np.dot(O_3x3_2x2, Mw), O_3x3_2x2.T )

def image_slice(x, X, B, D, pad=0):
    start = x * B - pad
    stop  = start + D
    pad = [0,0]
    if start < 0:
        pad[0] = -start
        start = 0
    if stop - 1 >= X:
        pad[1] = stop - X
    return start, stop, pad



def xprop_winograd(I, F, O, padding, minimal=False, backward=False):

    if backward:
        # C <=> K and mirror R,S
        F = np.transpose(F[:,::-1,::-1,:], (3,1,2,0)).copy()
        # Invert padding
        padding = [2 - p for p in padding]

    C, Y, X, N = I.shape
    K, P, Q, N = O.shape

    B = 2
    D = B + 2
    Yw = ceil_div(P, B)
    Xw = ceil_div(Q, B)

    Fw = np.empty((D,D,C,K))
    Iw = np.empty((D,D,C,Yw,Xw,N))
    Mw = np.empty((D,D,K,Yw,Xw,N)) #, dtype=np.float32

    # Transform Filters
    for c in range(C):
        for k in range(K):
            trans_F_2x2_3x3(Fw[:,:,c,k], F[c,:,:,k], minimal)

    # Iterate over image transform dimensions and slice out tiles of the image
    for y in range(Yw):
        start_y, stop_y, pad_y = image_slice(y, Y, B, D, padding[0])

        for x in range(Xw):
            start_x, stop_x, pad_x = image_slice(x, X, B, D, padding[1])

            sliceI = I[:, start_y:stop_y, start_x:stop_x, :]

            # add any zero padding if needed
            if any(pad_y) or any(pad_x):
                sliceI = np.pad(sliceI, ((0,0), pad_y, pad_x, (0,0)), 'constant')

            # Apply the Image transform
            for c in range(C):
                for n in range(N):
                    trans_I_2x2_3x3(Iw[:,:,c,y,x,n], sliceI[c,:,:,n], minimal)

    # Fw, scaleF = quantize(Fw, 8)
    # Iw, scaleI = quantize(Iw, 8)

    # Fw = Fw.astype(np.float32)
    # Iw = Iw.astype(np.float32)

    # Batched gemm for the pointwise multiplication step
    for s in range(D):
        for t in range(D):
            # [K,Yw,Xw,N] = [C,K].T . [C,YwXwN]
            Mw[s,t] = np.dot( Fw[s,t].T, Iw[s,t].reshape(C, -1) ).reshape((K,Yw,Xw,N))

    # Mw = Mw.astype(np.float32) * scaleF * scaleI

    # Mw = Mw.astype(np.float32)

    # Iterate over the convovled result in the pointwise space and apply inverse transform
    for y in range(Yw):
        p    = y * B
        plen = 2 if p + 1 < P else 1
        for x in range(Xw):
            q  = x * B
            qlen = 2 if q + 1 < Q else 1
            for k in range(K):
                for n in range(N):
                    # Toss out any points that don't fit
                    O[k,p:p + plen,q:q + qlen,n] = trans_O_2x2_3x3(Mw[:,:,k,y,x,n], minimal)[0:plen,0:qlen]



def updat_winograd(I, E, U, padding, minimal=False, inner=False):

    C, Y, X, N = I.shape
    K, P, Q, N = E.shape

    B = 2
    D = B + 2
    Yw = ceil_div(P, B)
    Xw = ceil_div(Q, B)

    Iw = np.empty((D,D,N,C))
    Ew = np.empty((D,D,N,K))
    if inner:
        Mw = np.empty((D,D,C,K))
        U.fill(0.0)
    else:
        Mw = np.zeros((D,D,C,K))

    for y in range(Yw):
        start_y, stop_y, pad_y = image_slice(y, Y, B, D, padding[0])
        start_p, stop_p, pad_p = image_slice(y, P, B, B)

        for x in range(Xw):
            start_x, stop_x, pad_x = image_slice(x, X, B, D, padding[1])
            start_q, stop_q, pad_q = image_slice(x, Q, B, B)

            sliceI = I[:, start_y:stop_y, start_x:stop_x, :]
            sliceE = E[:, start_p:stop_p, start_q:stop_q, :]

            if any(pad_y) or any(pad_x):
                sliceI = np.pad(sliceI, ((0,0), pad_y, pad_x, (0,0)), 'constant')

            if any(pad_p) or any(pad_q):
                sliceE = np.pad(sliceE, ((0,0), pad_p, pad_q, (0,0)), 'constant')

            for c in range(C):
                for n in range(N):
                    trans_I_3x3_2x2(Iw[:,:,n,c], sliceI[c,:,:,n], minimal)

            # print y,x
            # print Iw[:,:,0,0].reshape((16,))
            # exit()

            for k in range(K):
                for n in range(N):
                    trans_F_3x3_2x2(Ew[:,:,n,k], sliceE[k,:,:,n], minimal)

            # print Ew[:,:,0,0].reshape((16,))

            for s in range(D):
                for t in range(D):
                    # [C,K] += [N,C].T . [N,K]
                    if inner:
                        Mw[s,t]  = np.dot( Iw[s,t].T, Ew[s,t] )
                    else:
                        Mw[s,t] += np.dot( Iw[s,t].T, Ew[s,t] )

            # Transform can be applied in inner or outer loop
            if inner:
                for c in range(C):
                    for k in range(K):
                        U[c,:,:,k] += trans_O_3x3_2x2(Mw[:,:,c,k], minimal)

    # outer loop transform
    if not inner:
        for c in range(C):
            for k in range(K):
                U[c,:,:,k] = trans_O_3x3_2x2(Mw[:,:,c,k], minimal)



### Test Code ###

np.set_printoptions(threshold=8192 * 4, linewidth=600, formatter={'float':lambda x: "%6.3f" % x})

minimal = 1
ones = 0
N    = 32
C, K = 32, 32
Y, X = 4, 4
R, S = 3, 3     # Fixed winograd dim
strides = 1, 1  # Fixed winograd dim
padding = 1, 1  # 0-2

P = out_dim(R, Y, padding[0], strides[0])
Q = out_dim(S, X, padding[1], strides[1])

dimI = (C,Y,X,N)
dimF = (C,R,S,K)
dimO = (K,P,Q,N)

if ones:
    I  = np.ones(dimI)
    F  = np.ones(dimF)
    E  = np.ones(dimO)

    # for c in range(C):
    #     for n in range(N):
    #         I[c,:,:,n] = c+1 #np.arange(1+c,37+c).reshape((6,6))

    # for k in range(K):
    #     for n in range(N):
    #         E[k,:,:,n] = k+1 #np.arange(1+k,17+k).reshape((4,4))

else:
    I  = np.maximum(np.random.uniform(-1.0, 1.0, dimI), 0)
    F  = np.random.normal(0.0, 0.1, dimF)
    #F  = np.random.uniform(-1.0, 1.0, dimF)
    E  = np.random.uniform(-1.0, 1.0, dimO)

Od = np.empty(dimO)
Ow = np.empty(dimO) #, dtype=np.float32

Bd = np.empty(dimI)
Bw = np.empty(dimI) #, dtype=np.float32

Ud = np.empty(dimF)
Uw = np.empty(dimF)


xprop_direct(I, F, Od, padding, strides)
xprop_winograd(I, F, Ow, padding, minimal=minimal)

xprop_direct(E, F, Bd, padding, strides, backward=True)
xprop_winograd(E, F, Bw, padding, minimal=minimal, backward=True)

updat_direct(I, E, Ud, padding, strides)
updat_winograd(I, E, Uw, padding, minimal=minimal, inner=True)

difO = Od - Ow
difB = Bd - Bw
difU = Ud - Uw

neon_logger.display(abs(difO).max() / Od.max())
neon_logger.display(abs(difB).max() / Bd.max())
neon_logger.display(abs(difU).max() / Ud.max())

# print Bd[0,:,:,0]
# print Bw[0,:,:,0]

# print Ud[0,:,:,0]
# print Uw[0,:,:,0]
# print difU[0,:,:,0]



#!/usr/bin/env python
# Copyright 2016 Nervana Systems Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import division
from builtins import str
import numpy         as np
import pycuda.driver as drv
from neon import logger as neon_logger
from neon.backends.nervanagpu import NervanaGPU
from neon.backends.nervanacpu import NervanaCPU

from neon.backends.convolution import (_ceil_div,
    FpropCuda,   BpropCuda,   UpdateCuda,
    FpropDirect, BpropDirect, UpdateDirect)

from neon.backends.winograd_conv import (
    FpropWinograd_2x2_3x3, BpropWinograd_2x2_3x3, UpdateWinograd_3x3_2x2,
    FpropWinograd_4x4_3x3, BpropWinograd_4x4_3x3, UpdateWinograd_3x3_4x4,
    FpropWinograd_2x2_5x5, BpropWinograd_2x2_5x5)

fprop_kernels  = (FpropCuda,  FpropDirect,  FpropWinograd_2x2_3x3,  FpropWinograd_4x4_3x3, FpropWinograd_2x2_5x5)
bprop_kernels  = (BpropCuda,  BpropDirect,  BpropWinograd_2x2_3x3,  BpropWinograd_4x4_3x3, BpropWinograd_2x2_5x5)
update_kernels = (UpdateCuda, UpdateDirect, UpdateWinograd_3x3_2x2, UpdateWinograd_3x3_4x4)

ng = NervanaGPU(0)
nc = NervanaCPU()

neon_logger.display(drv.Context.get_current().get_device().name())

def main():
    out =  0
    ones = 0

    #                D,   H,   W,  T, R, S,    pad,   str
    conv_1x1     = ( 1,  14,  14,  1, 1, 1,  0,0,0, 1,1,1)
    conv_3x3     = ( 1,  14,  14,  1, 3, 3,  0,1,1, 1,1,1)
    conv_3x3p0   = ( 1,  14,  14,  1, 3, 3,  0,0,0, 1,1,1)
    conv_3x3p2   = ( 1,  14,  14,  1, 3, 3,  0,2,2, 1,1,1)
    conv_3x3s2   = ( 1,  14,  14,  1, 3, 3,  0,1,1, 1,2,2)
    conv_1x3     = ( 1,  14,  14,  1, 1, 3,  0,0,1, 1,1,1)
    conv_3x1     = ( 1,  14,  14,  1, 3, 1,  0,1,0, 1,1,1)
    conv_5x5     = ( 1,  14,  14,  1, 5, 5,  0,2,2, 1,1,1)
    conv_11x11s4 = ( 1, 224, 224,  1,11,11,  0,2,2, 1,4,4)
    conv_1x1x1   = ( 7,   7,   7,  1, 1, 1,  0,0,0, 1,1,1)
    conv_3x3x3   = ( 7,   7,   7,  3, 3, 3,  1,1,1, 1,1,1)
    conv_3x3x3s2 = ( 7,   7,   7,  3, 3, 3,  1,1,1, 2,2,2)
    conv_3x3L    = ( 1, 200, 200,  1, 3, 3,  0,1,1, 1,1,1)
    conv_1D      = ( 1, 13, 3263,  1,13,11,  0,0,0, 1,1,3)

    # configs = [

    # ]

    configs = [
        # Kernel                  N,   C,   K  Determ Cmpnd  Xtern, conv
        (FpropCuda,              32,  32,  32, True,  True,  None,  (conv_3x3,) ),
        (BpropCuda,              32,  32,  32, True,  True,  None,  (conv_3x3,) ),
        (UpdateCuda,             32,  32,  32, True,  True,  None,  (conv_3x3,) ),
        (FpropCuda,              32,  32,  32, True,  False, None,  (conv_1x1, conv_3x3s2, conv_1x3, conv_3x1, conv_5x5) ),
        (BpropCuda,              32,  32,  32, True,  False, None,  (conv_1x1, conv_3x3s2, conv_1x3, conv_3x1, conv_5x5) ),
        (UpdateCuda,             32,  32,  32, True,  False, None,  (conv_1x1, conv_3x3s2, conv_1x3, conv_3x1, conv_5x5) ),
        (FpropCuda,              32,   3,  64, True,  False, None,  (conv_11x11s4,) ),
        (UpdateCuda,             32,   3,  64, True,  False, None,  (conv_11x11s4,) ),

        (FpropDirect,            32,  32,  64, True,  True,  None,  (conv_3x3,conv_3x3L) ),
        (BpropDirect,            32,  64,  32, True,  True,  None,  (conv_3x3,conv_3x3L) ),
        (UpdateDirect,           32,  32,  32, True,  True,  None,  (conv_3x3,conv_3x3L) ),
        (UpdateDirect,           32,  32,  32, False, True,  None,  (conv_3x3,conv_3x3L) ),

        (FpropDirect,            32,  32,  64, True,  False, None,  (conv_1x1, conv_3x3s2, conv_1x3, conv_3x1, conv_5x5, conv_3x3x3, conv_1x1x1, conv_3x3x3s2) ),
        (BpropDirect,            32,  64,  32, True,  False, None,  (conv_1x1, conv_3x3s2, conv_1x3, conv_3x1, conv_5x5, conv_3x3x3, conv_1x1x1, conv_3x3x3s2) ),
        (UpdateDirect,           32,  32,  32, True,  False, None,  (conv_1x1, conv_3x3s2, conv_1x3, conv_3x1, conv_5x5, conv_3x3x3, conv_1x1x1, conv_3x3x3s2) ),

        (FpropDirect,            32,   3,  64, True,  False, None,  (conv_11x11s4,) ),
        (UpdateDirect,           32,   3,  32, True,  False, None,  (conv_11x11s4,) ),

        (FpropDirect,            32,  64, 128, True,  True,  None,  (conv_3x3,) ),
        (FpropDirect,            32,  32,  63, True,  True,  None,  (conv_3x3,conv_3x3L) ),
        (FpropDirect,            32,  32,   1, True,  True,  None,  (conv_3x3,conv_3x3L) ),

        (FpropDirect,            16,  32,  64, True,  False, None,  (conv_3x3,) ),
        (FpropDirect,             8,  32,  64, True,  False, None,  (conv_3x3,) ),
        (FpropDirect,             4,  32,  64, True,  False, None,  (conv_3x3,) ),
        (FpropDirect,             2,  32,  64, True,  False, None,  (conv_3x3,) ),
        (FpropDirect,             1,  32,  64, True,  True,  None,  (conv_3x3,) ),

        (UpdateDirect,           16,  32,  63, True,  False, None,  (conv_3x3,) ),
        (UpdateDirect,            8,  32,  64, True,  False, None,  (conv_3x3,) ),
        (UpdateDirect,            4,  32, 128, True,  False, None,  (conv_3x3,) ),

        (FpropDirect,            32,   1, 512, True,  False, None,  (conv_1D,) ),
        (FpropDirect,            16,   1, 512, True,  False, None,  (conv_1D,) ),
        (FpropDirect,             8,   1, 512, True,  False, None,  (conv_1D,) ),
        (FpropDirect,             4,   1, 512, True,  False, None,  (conv_1D,) ),
        (FpropDirect,             2,   1, 512, True,  False, None,  (conv_1D,) ),
        (FpropDirect,             1,   1, 512, True,  False, None,  (conv_1D,) ),

        (UpdateDirect,           32,   1, 512, True,  False, None,  (conv_1D,) ),
        (UpdateDirect,           16,   1, 512, True,  False, None,  (conv_1D,) ),
        (UpdateDirect,            8,   1, 512, True,  False, None,  (conv_1D,) ),
        (UpdateDirect,            4,   1, 512, True,  False, None,  (conv_1D,) ),

        # Kernel                  N,   C,   K  Determ Cmpnd  Xtern, conv
        (FpropDirect,            64,  32,  64, True,  True,  None,  (conv_3x3,) ),
        (FpropDirect,            64,  32, 128, True,  True,  None,  (conv_3x3,) ),
        (FpropDirect,           128,  32,  32, True,  True,  None,  (conv_3x3,) ),
        (FpropDirect,           128,  32,  64, True,  True,  None,  (conv_3x3,) ),
        (FpropDirect,           128,  32, 128, True,  True,  None,  (conv_3x3,) ),
        (BpropDirect,            64,  64,  32, True,  True,  None,  (conv_3x3,) ),
        (BpropDirect,            64, 128,  32, True,  True,  None,  (conv_3x3,) ),
        (BpropDirect,           128,  32,  32, True,  True,  None,  (conv_3x3,) ),
        (BpropDirect,           128,  64,  32, True,  True,  None,  (conv_3x3,) ),
        (BpropDirect,           128, 128,  32, True,  True,  None,  (conv_3x3,) ),

        (FpropDirect,            64,  32,  64, True,  False, None,  (conv_1x1, conv_3x3s2, conv_1x3, conv_3x1, conv_5x5, conv_3x3x3, conv_1x1x1, conv_3x3x3s2) ),
        (FpropDirect,            64,  32, 128, True,  False, None,  (conv_1x1, conv_3x3s2, conv_1x3, conv_3x1, conv_5x5, conv_3x3x3, conv_1x1x1, conv_3x3x3s2) ),
        (FpropDirect,           128,  32,  32, True,  False, None,  (conv_1x1, conv_3x3s2, conv_1x3, conv_3x1, conv_5x5, conv_3x3x3, conv_1x1x1, conv_3x3x3s2) ),
        (FpropDirect,           128,  32,  64, True,  False, None,  (conv_1x1, conv_3x3s2, conv_1x3, conv_3x1, conv_5x5, conv_3x3x3, conv_1x1x1, conv_3x3x3s2) ),
        (FpropDirect,           128,  32, 128, True,  False, None,  (conv_1x1, conv_3x3s2, conv_1x3, conv_3x1, conv_5x5, conv_3x3x3, conv_1x1x1, conv_3x3x3s2) ),
        (BpropDirect,            64,  64,  32, True,  False, None,  (conv_1x1, conv_3x3s2, conv_1x3, conv_3x1, conv_5x5, conv_3x3x3, conv_1x1x1, conv_3x3x3s2) ),
        (BpropDirect,            64, 128,  32, True,  False, None,  (conv_1x1, conv_3x3s2, conv_1x3, conv_3x1, conv_5x5, conv_3x3x3, conv_1x1x1, conv_3x3x3s2) ),
        (BpropDirect,           128,  32,  32, True,  False, None,  (conv_1x1, conv_3x3s2, conv_1x3, conv_3x1, conv_5x5, conv_3x3x3, conv_1x1x1, conv_3x3x3s2) ),
        (BpropDirect,           128,  64,  32, True,  False, None,  (conv_1x1, conv_3x3s2, conv_1x3, conv_3x1, conv_5x5, conv_3x3x3, conv_1x1x1, conv_3x3x3s2) ),
        (BpropDirect,           128, 128,  32, True,  False, None,  (conv_1x1, conv_3x3s2, conv_1x3, conv_3x1, conv_5x5, conv_3x3x3, conv_1x1x1, conv_3x3x3s2) ),

        (FpropDirect,            64,   3,  64, True,  False, None,  (conv_11x11s4,) ),
        (FpropDirect,            64,   3, 128, True,  False, None,  (conv_11x11s4,) ),
        (FpropDirect,           128,   3,  32, True,  False, None,  (conv_11x11s4,) ),
        (FpropDirect,           128,   3,  64, True,  False, None,  (conv_11x11s4,) ),

        (FpropDirect,            64,  33,  56, True,  True,  None,  (conv_3x3s2,) ),
        (FpropDirect,            64,  33, 120, True,  True,  None,  (conv_3x3s2,) ),
        (FpropDirect,           128,  33,  56, True,  True,  None,  (conv_3x3s2,) ),
        (FpropDirect,           128,  33, 120, True,  True,  None,  (conv_3x3s2,) ),
        (FpropDirect,           128,  33, 248, True,  True,  None,  (conv_3x3s2,) ),

        # Kernel                  N,   C,   K  Determ Cmpnd  Xtern, conv
        (FpropWinograd_2x2_3x3,  32,  32,  32, True,  True,  False, (conv_3x3,conv_3x3L) ),
        (FpropWinograd_2x2_3x3,  32,  32,  32, True,  True,  True,  (conv_3x3,) ),
        (BpropWinograd_2x2_3x3,  32,  32,  32, True,  True,  False, (conv_3x3,conv_3x3L) ),
        (BpropWinograd_2x2_3x3,  32,  32,  32, True,  True,  True,  (conv_3x3,) ),
        (UpdateWinograd_3x3_2x2, 32,  32,  32, True,  True,  None,  (conv_3x3,) ),
        (UpdateWinograd_3x3_2x2, 32,  32,  32, False, True,  None,  (conv_3x3,) ),
        (FpropWinograd_4x4_3x3,  32,  32,  32, True,  True,  False, (conv_3x3,) ),
        (FpropWinograd_4x4_3x3,  32,  32,  32, True,  True,  True,  (conv_3x3,conv_3x3L) ),
        (BpropWinograd_4x4_3x3,  32,  32,  32, True,  True,  False, (conv_3x3,) ),
        (BpropWinograd_4x4_3x3,  32,  32,  32, True,  True,  True,  (conv_3x3,conv_3x3L) ),
        (UpdateWinograd_3x3_4x4, 32,  32,  32, True,  True,  None,  (conv_3x3,) ),
        (UpdateWinograd_3x3_4x4, 32,  32,  32, False, True,  None,  (conv_3x3,) ),

        (FpropWinograd_2x2_3x3,  32,  32,  32, True,  False, True,  (conv_3x3p0,conv_3x3p2) ),
        (BpropWinograd_2x2_3x3,  32,  32,  32, True,  False, True,  (conv_3x3p0,conv_3x3p2) ),
        (UpdateWinograd_3x3_2x2, 32,  32,  32, True,  False, None,  (conv_3x3p0,conv_3x3p2) ),
        (FpropWinograd_4x4_3x3,  32,  32,  32, True,  False, True,  (conv_3x3p0,conv_3x3p2) ),
        (BpropWinograd_4x4_3x3,  32,  32,  32, True,  False, True,  (conv_3x3p0,conv_3x3p2) ),
        (UpdateWinograd_3x3_4x4, 32,  32,  32, True,  False, None,  (conv_3x3p0,conv_3x3p2) ),

        (FpropWinograd_2x2_3x3,   1,  63,  63, True,  False, True,  (conv_3x3,conv_3x3L) ),
        (BpropWinograd_2x2_3x3,   1,  63,  63, True,  False, True,  (conv_3x3,conv_3x3L) ),
        (UpdateWinograd_3x3_2x2,  1,  63,  63, True,  False, None,  (conv_3x3,) ),
        (FpropWinograd_4x4_3x3,   1,  63,  63, True,  False, True,  (conv_3x3,conv_3x3L) ),
        (BpropWinograd_4x4_3x3,   1,  63,  63, True,  False, True,  (conv_3x3,conv_3x3L) ),
        (UpdateWinograd_3x3_4x4,  1,  63,  63, True,  False, None,  (conv_3x3,) ),

        (FpropWinograd_2x2_5x5,  32,  32,  32, False, True,  None,  (conv_5x5,) ),
        (BpropWinograd_2x2_5x5,  32,  32,  32, False, True,  None,  (conv_5x5,) ),

        (FpropWinograd_2x2_5x5,  32,  64, 192, False, False, None,  (conv_5x5,) ),
        (BpropWinograd_2x2_5x5,  32,  64, 192, False, False, None,  (conv_5x5,) ),
        (FpropWinograd_2x2_5x5,  16,  64, 192, False, False, None,  (conv_5x5,) ),
        (FpropWinograd_2x2_5x5,   8,  64, 192, False, False, None,  (conv_5x5,) ),
        (FpropWinograd_2x2_5x5,   4,  64, 192, False, False, None,  (conv_5x5,) ),
        (FpropWinograd_2x2_5x5,   2,  64, 192, False, False, None,  (conv_5x5,) ),
    ]

    fprop_opts = [
        dict(),
        dict(slope=0.0, relu=True),
        dict(slope=0.1, relu=True),
        dict(bias=True),
        dict(bias=True, slope=0.0, relu=True),
        dict(bias=True, slope=0.1, relu=True),
        dict(bsum=True),
    ]
    bprop_opts = [
        dict(),
        dict(X=True, slope=0.0, brelu=True),
        dict(X=True, slope=0.1, brelu=True),
        dict(X=True, bsum=True, slope=0.0, brelu=True),
        dict(X=True, bsum=True, slope=0.1, brelu=True),
        dict(X=True, alpha=2.0, beta=3.0),
        dict(alpha=2.0, beta=3.0),
    ]
    update_opts = [
        dict(alpha=2.0, beta=3.0),
        dict(),
    ]

    for config in configs:

        kernelClass, N, C, K, determ, compound, override, convs = config

        for conv in convs:

            D, H, W, T, R, S, pad_d, pad_h, pad_w, str_d, str_h, str_w = conv

            ng.deterministic = determ

            layer = nc.conv_layer(np.float64,
                N, C, K, D, H, W, T, R, S,
                pad_d, pad_h, pad_w,
                str_d, str_h, str_w)

            (M, P, Q) = layer.MPQ

            if kernelClass in (FpropCuda, BpropCuda, UpdateCuda):
                dtypes = (np.float32,)
            else:
                dtypes = (np.float32, np.float16)

            for dtype in (dtypes):

                ng.scratch_buffer_reset()

                if override is None:
                    kernel = kernelClass(ng, np.dtype(dtype),
                        N, C, K, D, H, W, T, R, S, M, P, Q,
                        pad_d, pad_h, pad_w,
                        str_d, str_h, str_w)
                else:
                    kernel = kernelClass(ng, np.dtype(dtype),
                        N, C, K, D, H, W, T, R, S, M, P, Q,
                        pad_d, pad_h, pad_w,
                        str_d, str_h, str_w, override)

                neon_logger.display(kernel)

                back = False
                if kernelClass in fprop_kernels:
                    dimI1 = layer.dimI
                    dimI2 = layer.dimF
                    dimO  = layer.dimO
                    opts  = fprop_opts
                    func  = layer.xprop_conv
                elif kernelClass in bprop_kernels:
                    dimI1 = layer.dimO
                    dimI2 = layer.dimF
                    dimO  = layer.dimI
                    opts  = bprop_opts
                    func  = layer.xprop_conv
                    back  = True
                elif kernelClass in update_kernels:
                    dimI1 = layer.dimI
                    dimI2 = layer.dimO
                    dimO  = layer.dimF
                    opts  = update_opts
                    func  = layer.update_conv
                else:
                    raise TypeError("Unknown Kernel Class")

                if not compound:
                    opts = [ dict() ]

                if ones:
                    vals = 1.0
                else:
                    vals = (0.5 - ng.rand()) * 2

                devI1    = ng.empty(dimI1, dtype=dtype)
                devI2    = ng.empty(dimI2, dtype=dtype)
                devO     = ng.empty(dimO,  dtype=dtype)
                devI1[:] = vals
                devI2[:] = vals
                devO[:]  = vals
                cpuI1    = nc.array(devI1.get(), dtype=np.float64)
                cpuI2    = nc.array(devI2.get(), dtype=np.float64)
                cpuO     = nc.array(devO.get(),  dtype=np.float64)

                if compound and opts is not update_opts:
                    devB    = ng.empty((dimO[0], 1), dtype=np.float32)
                    devS    = ng.empty((dimO[0], 1), dtype=np.float32)
                    devB[:] = vals
                    devS[:] = vals
                    cpuB    = nc.array(devB.get(), dtype=np.float64)
                    cpuS    = nc.array(devS.get(), dtype=np.float64)

                if opts is bprop_opts:
                    devX    = ng.empty(dimO,  dtype=dtype)
                    devX[:] = vals
                    cpuX    = nc.array(devX.get(),  dtype=np.float64)

                for opt in opts:
                    dev_opts = dict(opt)
                    cpu_opts = dict(opt)

                    if back:
                        cpu_opts["backward"] = True

                    if "bias" in dev_opts:
                        dev_opts["bias"] = devB
                        cpu_opts["bias"] = cpuB
                    if "bsum" in dev_opts:
                        dev_opts["bsum"] = devS
                        cpu_opts["bsum"] = cpuS
                    if "X" in dev_opts:
                        dev_opts["X"] = devX
                        cpu_opts["X"] = cpuX

                    kernel.bind_params(devI1, devI2, devO, **dev_opts)
                    kernel.execute()

                    func(cpuI1, cpuI2, cpuO, **cpu_opts)

                    devA = devO.get()
                    cpuA = cpuO._tensor
                    difA = cpuA - devA

                    if out:
                        np.savetxt("out.txt",  difA.reshape((-1,dimO[-1])), fmt='%6.3f')
                        np.savetxt("outC.txt", cpuA.reshape((-1,dimO[-1])), fmt='%6.3f')
                        np.savetxt("outD.txt", devA.reshape((-1,dimO[-1])), fmt='%6.3f')

                    maxval = abs(cpuA).max()
                    maxdif = abs(difA).max()
                    ratio  = maxdif / maxval

                    if "bsum" in dev_opts:
                        devZ = devS.get()
                        cpuZ = cpuS._tensor
                        difZ = abs(cpuZ - devZ) / abs(cpuZ).max()
                        ratio2  = difZ.max()
                        #print difZ

                        # def output_slice(p, P, B):
                        #     p0 = p * B
                        #     p1 = p0 + B
                        #     if p1 > P:
                        #         p1 = P
                        #     return slice(p0, p1)

                        # B = 4
                        # Yw = _ceil_div(P, B)
                        # Xw = _ceil_div(Q, B)

                        # bsum = np.empty((K, Yw, Xw))

                        # for y in range(Yw):
                        #     pSlice = output_slice(y, P, B)
                        #     for x in range(Xw):
                        #         qSlice = output_slice(x, Q, B)
                        #         for k in range(K):
                        #             bsum[k,y,x] = np.sum(cpuA[k,0,pSlice,qSlice,:])

                        # bsum = bsum.reshape((K,-1))

                        # np.savetxt("outC.txt", bsum, fmt='%6.1f')
                        # np.savetxt("outD.txt", kernel.bsum.ary, fmt='%6.1f')
                        # np.savetxt("out.txt",  abs(bsum - kernel.bsum.ary), fmt='%6.1f')

                        #print abs(devZ - np.sum(kernel.bsum.ary, axis=1, keepdims=True))


                    else:
                        ratio2  = 0.0

                    bad = ratio > 0.01 or ratio2 > 0.01

                    if bad:
                        neon_logger.display("=================FAIL==============")

                    neon_logger.display("%17.12f %17.12f %s" % (ratio, ratio2, str(opt)))

                    if bad: exit()

                devI1 = devI2 = devO = devB = devS = devX = None
                cpuI1 = cpuI2 = cpuO = cpuB = cpuS = cpuX = None


if __name__ == '__main__':
    main()
